"""Functions for working with Protein Structure Graphs. Based on tests written by Eric Ma in PIN Library"""

import copy
from functools import partial

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from pathlib import Path

import networkx as nx
import pandas as pd
import pytest

import graphein.protein as gp
from graphein.protein.config import ProteinGraphConfig
from graphein.protein.edges.distance import (
    INFINITE_DIST,
    add_k_nn_edges,
    add_salt_bridges,
    add_sequence_distance_edges,
    add_vdw_clashes,
    add_vdw_interactions,
    filter_distmat,
    get_edges_by_bond_type,
    get_ring_atoms,
    get_ring_centroids,
)
from graphein.protein.graphs import construct_graph, read_pdb_to_dataframe
from graphein.protein.resi_atoms import (
    AROMATIC_RESIS,
    BOND_TYPES,
    CATION_RESIS,
    HYDROPHOBIC_RESIS,
    NEG_AA,
    PI_RESIS,
    POS_AA,
    RESI_NAMES,
    SALT_BRIDGE_ANIONS,
    SALT_BRIDGE_CATIONS,
    SULPHUR_RESIS,
    VDW_RADII,
)

DATA_PATH = Path(__file__).resolve().parent.parent / "test_data" / "4hhb.pdb"


def generate_graph():
    """Generate PDB network.
    This is a helper function.
    """
    return construct_graph(path=str(DATA_PATH))


@pytest.fixture(scope="module")
def net():
    """Generate proteingraph from 2VUI.pdb."""
    return generate_graph()


@pytest.fixture()
def pdb_df():
    """Generate pdb_df from 2VIU.pdb."""
    return read_pdb_to_dataframe(DATA_PATH)


def test_nodes_are_strings(net):
    """
    Checks to make sure that the nodes are a string.
    For expediency, checks only 1/4 of the nodes.
    """
    for n in net.nodes():
        assert isinstance(n, str)


def test_add_hydrophobic_interactions(net):
    """Test the function add_hydrophobic_interactions_."""
    resis = get_edges_by_bond_type(net, "hydrophobic")
    for r1, r2 in resis:
        assert net.nodes[r1]["residue_name"] in HYDROPHOBIC_RESIS
        assert net.nodes[r2]["residue_name"] in HYDROPHOBIC_RESIS


def test_add_disulfide_interactions(net):
    """Test the function add_disulfide_interactions_."""
    resis = get_edges_by_bond_type(net, "disulfide")

    for r1, r2 in resis:
        assert net.nodes[r1]["residue_name"] == "CYS"
        assert net.nodes[r2]["residue_name"] == "CYS"


@pytest.mark.skip(reason="Not yet implemented.")
def test_delaunay_triangulation(net):
    """
    Test delaunay triangulation.
    I am including this test here that always passes because I don't know how
    best to test it. The code in pin.py uses scipy's delaunay triangulation.
    """
    pass


@pytest.mark.skip(reason="Implementation needs to be checked.")
def test_add_hydrogen_bond_interactions(net):
    """Test that the addition of hydrogen bond interactions works correctly."""
    pass


def test_add_aromatic_interactions(net):
    """
    Tests the function add_aromatic_interactions_.
    The test checks that each residue in an aromatic interaction
    is one of the aromatic residues.
    """
    resis = get_edges_by_bond_type(net, "aromatic")
    for n1, n2 in resis:
        assert net.nodes[n1]["residue_name"] in AROMATIC_RESIS
        assert net.nodes[n2]["residue_name"] in AROMATIC_RESIS


def test_add_aromatic_sulphur_interactions(net):
    """Tests the function add_aromatic_sulphur_interactions_."""
    resis = get_edges_by_bond_type(net, "aromatic_sulphur")
    for n1, n2 in resis:
        condition1 = (
            net.nodes[n1]["residue_name"] in SULPHUR_RESIS
            and net.nodes[n2]["residue_name"] in AROMATIC_RESIS
        )

        condition2 = (
            net.nodes[n2]["residue_name"] in SULPHUR_RESIS
            and net.nodes[n1]["residue_name"] in AROMATIC_RESIS
        )

        assert condition1 or condition2


def test_add_ionic_interactions(net):
    """
    Tests the function add_ionic_interactions_.
    This test checks that residues involved in ionic interactions
    are indeed oppositely-charged.
    Another test is needed to make sure that ionic interactions
    are not missed.
    """
    resis = get_edges_by_bond_type(net, "ionic")
    for n1, n2 in resis:
        resi1 = net.nodes[n1]["residue_name"]
        resi2 = net.nodes[n2]["residue_name"]

        condition1 = resi1 in POS_AA and resi2 in NEG_AA
        condition2 = resi2 in POS_AA and resi1 in NEG_AA

        assert condition1 or condition2


@pytest.mark.skip(reason="Not yet implemented.")
def test_get_ring_centroids(pdb_df):
    """Test the function get_ring_centroids."""
    print(pdb_df)
    ring_atom_TYR = get_ring_atoms(pdb_df, "TYR")
    assert len(ring_atom_TYR) == 32
    centroid_TYR = get_ring_centroids(ring_atom_TYR)
    assert len(centroid_TYR) == 16

    ring_atom_PHE = get_ring_atoms(pdb_df, "PHE")
    assert len(ring_atom_PHE) == 36
    centroid_PHE = get_ring_centroids(ring_atom_PHE)
    assert len(centroid_PHE) == 18


def test_add_cation_pi_interactions(net):
    """Tests the function add_cation_pi_interactions."""
    resis = get_edges_by_bond_type(net, "cation_pi")
    for n1, n2 in resis:
        resi1 = net.nodes[n1]["residue_name"]
        resi2 = net.nodes[n2]["residue_name"]

        condition1 = resi1 in CATION_RESIS and resi2 in PI_RESIS
        condition2 = resi2 in CATION_RESIS and resi1 in PI_RESIS

        assert condition1 or condition2


def test_add_peptide_bonds():
    file_path = Path(__file__).parent.parent / "test_data/4hhb.pdb"
    # Peptide bonds are default
    G = construct_graph(path=str(file_path))

    for u, v in G.edges():
        assert abs(int(u.split(":")[2]) - int(v.split(":")[2])) == 1


def test_add_sequence_distance_edges():
    D = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    file_path = Path(__file__).parent.parent / "test_data/4hhb.pdb"

    for d in D:
        config = ProteinGraphConfig(
            edge_construction_functions=[
                partial(add_sequence_distance_edges, d=d)
            ]
        )
        G = construct_graph(path=str(file_path), config=config)
        for u, v in G.edges():
            assert abs(int(u.split(":")[2]) - int(v.split(":")[2])) == d


def test_salt_bridge_interactions():
    """
    Test to ensure salt bridges are only between anionic and cationic residues.
    Also checks distances are correct on atom graphs.
    """
    g = construct_graph(
        pdb_code="216L",
        config=ProteinGraphConfig(
            edge_construction_functions=[add_salt_bridges]
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "salt_bridge" in d["kind"]
        res1 = g.nodes[u]["residue_name"]
        res2 = g.nodes[v]["residue_name"]

        assert (
            res1 in SALT_BRIDGE_CATIONS and res2 in SALT_BRIDGE_ANIONS
        ) or (
            res1 in SALT_BRIDGE_ANIONS and res2 in SALT_BRIDGE_CATIONS
        ), "Salt bridge not between cations and anions"

    # Test distances on atom graphs
    g = construct_graph(
        pdb_code="216L",
        config=ProteinGraphConfig(
            edge_construction_functions=[add_salt_bridges], granularity="atom"
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "salt_bridge" in d["kind"]
        res1 = g.nodes[u]["residue_name"]
        res2 = g.nodes[v]["residue_name"]

        assert (
            res1 in SALT_BRIDGE_CATIONS and res2 in SALT_BRIDGE_ANIONS
        ) or (
            res1 in SALT_BRIDGE_ANIONS and res2 in SALT_BRIDGE_CATIONS
        ), "Salt bridge not between cations and anions"
        assert d["distance"] < 4.0, "Salt bridge distance is too long"


def test_vdw_interactions():
    """
    Test to ensure vdw interactions are identified and are of correct length in
    atomic graphs.
    """
    g = construct_graph(
        pdb_code="216L",
        config=ProteinGraphConfig(
            edge_construction_functions=[add_vdw_interactions]
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "vdw" in d["kind"], "Edge not vdw"

    # Test distances on atom graphs
    g = construct_graph(
        pdb_code="216L",
        config=ProteinGraphConfig(
            edge_construction_functions=[add_vdw_interactions],
            granularity="atom",
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "vdw" in d["kind"]
        res1 = g.nodes[u]["element_symbol"]
        res2 = g.nodes[v]["element_symbol"]
        rad1 = VDW_RADII[res1]
        rad2 = VDW_RADII[res2]

        assert res1 != "H" and res2 != "H", "Hydrogen in vdw interaction"
        assert (
            d["distance"] < 0.5 + rad1 + rad2
        ), f"Vdw distance is too long: {d['distance']}"


def test_vdw_clashes():
    """
    Test to ensure vdw clashes are identified and are of correct length in
    atomic graphs.
    """
    g = construct_graph(
        pdb_code="216L",
        config=ProteinGraphConfig(
            edge_construction_functions=[add_vdw_clashes]
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "vdw_clash" in d["kind"], "Edge not vdw_clash"

    # Test distances on atom graphs
    g = gp.construct_graph(
        pdb_code="216L",
        config=gp.ProteinGraphConfig(
            edge_construction_functions=[add_vdw_clashes],
            granularity="atom",
        ),
    )
    for u, v, d in g.edges(data=True):
        assert "vdw_clash" in d["kind"]
        res1 = g.nodes[u]["element_symbol"]
        res2 = g.nodes[v]["element_symbol"]
        rad1 = VDW_RADII[res1]
        rad2 = VDW_RADII[res2]

        assert res1 != "H" and res2 != "H", "Hydrogen in vdw clash interaction"
        assert (
            d["distance"] < rad1 + rad2
        ), f"Vdw clash distance is too long: {d['distance']}"


def test_filter_distmat():
    pdb_df = pd.DataFrame(
        {
            "node_id": ["A:HIS:1", "A:TYR:2", "B:ALA:3"],
            "chain_id": ["A", "A", "B"],
        }
    )

    distmat = pd.DataFrame([[0.0, 1.0, 2.0], [1.0, 0.0, 2.0], [2.0, 2.0, 0.0]])

    exclude_edges_vals = [["intra"], ["inter"]]
    expected_distmats = [
        pd.DataFrame(
            [
                [0.0, INFINITE_DIST, 2.0],
                [INFINITE_DIST, 0.0, 2.0],
                [2.0, 2.0, 0.0],
            ]
        ),
        pd.DataFrame(
            [
                [0.0, 1.0, INFINITE_DIST],
                [1.0, 0.0, INFINITE_DIST],
                [INFINITE_DIST, INFINITE_DIST, 0.0],
            ]
        ),
    ]

    for vals, distmat_expected in zip(exclude_edges_vals, expected_distmats):
        distmat_out = filter_distmat(pdb_df, distmat, vals, False)
        pd.testing.assert_frame_equal(distmat_expected, distmat_out)


def test_add_k_nn_edges():
    pdb_df = pd.DataFrame(
        {
            "residue_number": [1, 2, 3, 4],
            "node_id": ["A:HIS:1", "A:TYR:2", "B:ALA:3", "B:ALA:4"],
            "chain_id": ["A", "A", "B", "B"],
            "x_coord": [1.0, 2.0, 4.0, 8.0],
            "y_coord": [0.0, 0.0, 0.0, 0.0],
            "z_coord": [0.0, 0.0, 0.0, 0.0],
        }
    )
    g_empty = nx.empty_graph(pdb_df["node_id"])
    g_empty.graph["pdb_df"] = pdb_df
    args_edges_pairs = [
        (
            dict(k=1),
            [
                ("A:HIS:1", "A:TYR:2"),
                ("A:TYR:2", "B:ALA:3"),
                ("B:ALA:3", "B:ALA:4"),
            ],
        ),
        (
            dict(k=2),
            [
                ("A:HIS:1", "A:TYR:2"),
                ("A:HIS:1", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:4"),
                ("B:ALA:3", "B:ALA:4"),
            ],
        ),
        (
            dict(k=1, exclude_edges=["intra"]),
            [
                ("A:HIS:1", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:4"),
            ],
        ),
        (
            dict(k=1, exclude_edges=["inter"]),
            [
                ("A:HIS:1", "A:TYR:2"),
                ("B:ALA:3", "B:ALA:4"),
            ],
        ),
        (
            dict(k=2, exclude_edges=["intra"], exclude_self_loops=False),
            [
                ("A:HIS:1", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:3"),
                ("A:TYR:2", "B:ALA:4"),
                ("A:HIS:1", "A:HIS:1"),
                ("A:TYR:2", "A:TYR:2"),
                ("B:ALA:3", "B:ALA:3"),
                ("B:ALA:4", "B:ALA:4"),
            ],
        ),
        (
            dict(k=1, exclude_edges=["intra"], exclude_self_loops=False),
            [
                ("A:HIS:1", "A:HIS:1"),
                ("A:TYR:2", "A:TYR:2"),
                ("B:ALA:3", "B:ALA:3"),
                ("B:ALA:4", "B:ALA:4"),
            ],
        ),
    ]

    for args, edges_expected in args_edges_pairs:
        g = copy.deepcopy(g_empty)
        add_k_nn_edges(g, **args)
        edges_real = list(g.edges())
        assert set(edges_real) == set(edges_expected)


def test_insertion_codes_in_add_k_nn_edges():
    pdb_df = pd.DataFrame(
        {
            "residue_number": [1, 2, 3],
            "node_id": ["A:HIS:1", "A:TYR:2", "B:ALA:4:altB"],
            "chain_id": ["A", "A", "B"],
            "x_coord": [1.0, 2.0, 3.0],
            "y_coord": [4.0, 5.0, 6.0],
            "z_coord": [7.0, 8.0, 9.0],
        },
        index=[0, 1, 3],  # simulating dropped "B:ALA:4:altA"
    )
    g = nx.empty_graph(pdb_df["node_id"])
    g.graph["pdb_df"] = pdb_df
    add_k_nn_edges(g, k=3)

    edges_expected = [
        ("A:HIS:1", "A:TYR:2"),
        ("A:HIS:1", "B:ALA:4:altB"),
        ("A:TYR:2", "B:ALA:4:altB"),
    ]

    assert set(g.edges()) == set(edges_expected)



from .pdb_data import PDBManager

try:
    from .torch_geometric_dataset import (
        InMemoryProteinGraphDataset,
        ProteinGraphDataset,
        ProteinGraphListDataset,
    )
except (NameError, ImportError):
    pass



"""Tests for graphein.testing.utils"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from graphein.protein import construct_graph
from graphein.testing.utils import (
    compare_approximate,
    compare_exact,
    edge_data_equal,
    edges_equal,
    graphs_equal,
    graphs_isomorphic,
    node_data_equal,
    nodes_equal,
)


def test_graph_isomorphism():
    """Test that the graph isomorphism is working."""
    file_path = Path(__file__).parent.parent / "protein" / "test_data/4hhb.pdb"
    G_1 = construct_graph(path=str(file_path))
    G_2 = construct_graph(path=str(file_path))
    assert graphs_isomorphic(G_1, G_2), "Graphs are not isomorphic."
    assert graphs_equal(G_1, G_2), "Graphs are not equal."
    assert nodes_equal(G_1, G_2), "Graphs do not contain the same nodes."
    assert node_data_equal(G_1, G_2), "Node features differ."
    assert edges_equal(G_1, G_2), "Graphs do not contain the same edges."
    assert edge_data_equal(G_1, G_2), "Edge features differ."

    file_path = Path(__file__).parent.parent / "protein" / "test_data/1lds.pdb"
    G_2 = construct_graph(path=str(file_path))
    assert not graphs_isomorphic(G_1, G_2), "Graphs not isomorphic."
    assert not graphs_equal(G_1, G_2), "Graphs are equal."
    assert not nodes_equal(G_1, G_2), "Graphs contain the same nodes."
    assert not node_data_equal(G_1, G_2), "Node features do not differ."
    assert not edges_equal(G_1, G_2), "Graphs contain the same edges."
    assert not edge_data_equal(G_1, G_2), "Edge features do not differ."


def test_dictionaries_equal():
    dict_1 = {
        "a": 1,
        "b": 2,
        "c": np.array([0, 1, 2]),
        "d": pd.DataFrame({"a": [1, 2, 3]}),
    }
    dict_2 = {
        "a": 1,
        "b": 2,
        "c": np.array([0, 1, 3]),
        "d": pd.DataFrame({"a": [1, 2, 3]}),
    }
    dict_3 = {
        "a": 1,
        "b": 2,
        "c": np.array([0, 1, 2]),
        "d": pd.DataFrame({"a": [1, 2, 4]}),
    }

    assert compare_exact(dict_1, dict_1), "Dictionaries are not equal."
    assert compare_approximate(dict_1, dict_1), "Dictionaries are not equal."

    assert not compare_exact(dict_1, dict_2), "Dictionaries are equal."
    assert not compare_exact(dict_1, dict_3), "Dictionaries are equal."
    assert not compare_exact(dict_2, dict_3), "Dictionaries are equal."

    assert not compare_approximate(dict_1, dict_2), "Dictionaries are equal."
    assert not compare_approximate(dict_1, dict_3), "Dictionaries are equal."
    assert not compare_approximate(dict_2, dict_3), "Dictionaries are equal."



"""Contains utilities for computing analytics on and plotting summaries of Protein Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import itertools
from collections import Counter
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd
import plotly as ply
import plotly.express as px


def plot_degree_distribution(
    g: nx.Graph, title: Optional[str] = None
) -> plotly.graph_objects.Figure:
    """Plots the distribution of node degrees in the graph.

    :param g: networkx graph to plot the distribution of node degrees in.
    :type g: nx.Graph
    :param title: Title of plot. defaults to ``None``.
    :type title: Optional[str], optional
    :return: Plotly figure.
    :rtpe: plotly.graph_objects.Figure
    """
    if not title:
        title = g.graph["name"] + " - Degree Distribution"
    df = pd.DataFrame(g.degree())
    df.columns = ["node", "degree"]
    return ply.hist_frame(df, x="degree", title=title)


def plot_residue_composition(
    g: nx.Graph, sort_by: Optional[str] = None, plot_type: str = "bar"
) -> plotly.graph_objects.Figure:
    """Plots the residue composition of the graph.

    :param g: Protein graph to plot the residue composition of.
    :type g: nx.Graph
    :param sort_by: How to sort the values (``"alphabetical"``, ``"count"``), defaults to ``None`` (no sorting).
    :type sort_by: Optional[str], optional
    :param plot_type: How to plot the composition (``"bar"``, ``"pie"``), defaults to ``"bar"``.
    :type plot_type: str, optional
    :raises ValueError: Raises ValueError if ``sort_by`` is not one of ``"alphabetical"``, ``"count"``.
    :return: Plotly figure.
    :rtype: plotly.graph_objects.Figure
    """
    title = g.graph["name"] + " - Residue Composition"
    residues = [d["residue_name"] for _, d in g.nodes(data=True)]
    counts = pd.Series(Counter(residues))
    if not sort_by:
        counts = pd.DataFrame(counts)
    elif sort_by == "alphabetical":
        counts = pd.DataFrame(counts).sort_index()
    elif sort_by == "count":
        counts = pd.DataFrame(counts).sort_values(by=0, ascending=False)
    else:
        raise ValueError(
            f"sort_by: {sort_by} not supported. Please use on of 'count' or 'alphabetical'."
        )
    counts.columns = ["counts"]
    if plot_type == "bar":
        fig = px.bar(counts, title=title)
    elif plot_type == "pie":
        fig = px.pie(counts, values="counts", names=counts.index, title=title)
    return fig


def plot_degree_by_residue_type(
    g: nx.Graph, normalise_by_residue_occurrence: bool = True
) -> plotly.graph_objects.Figure:
    """Plots the distribution of node degrees in the graph.

    :param g: networkx graph to plot the distribution of node degrees by residue type of.
    :type g: nx.Graph
    :param normalise_by_residue_occurrence: Whether to normalise the degree by the number of residues of the same type.
    :type normalise_by_residue_occurrence: bool
    :return: Plotly figure.
    :rtpe: plotly.graph_objects.Figure
    """

    title = g.graph["name"] + " - Total Degree by Residue Type"

    residues = [d["residue_name"] for _, d in g.nodes(data=True)]
    counts = pd.Series(Counter(residues))
    residues = list(set(residues))

    degree_values = {r: 0 for r in residues}
    for r, deg in g.degree():
        degree_values[r[2:5]] += deg

    df = pd.Series(degree_values, index=residues)
    if normalise_by_residue_occurrence:
        df = df.divide(counts)
        title += " (Normalised by Residue Occurrence)"

    return ply.hist_frame(df, x=df.index, y=df.values, title=title)


def plot_edge_type_distribution(
    g: nx.Graph, plot_type: str = "bar", title: Optional[str] = None
) -> plotly.graph_objects.Figure:
    """Plots the distribution of edge types in the graph.

    :param g: NetworkX graph to plot the distribution of edge types in.
    :type g: nx.Graph
    :param plot_type: Type of plot to produce, defaults to ``"bar"``. One of ``"bar"``, ``"pie"``.
    :type plot_type: str, optional
    :param title: Title of plot. defaults to None
    :type title: Optional[str], optional
    :return: Plotly figure.
    :rtype: plotly.graph_objects.Figure
    """
    if not title:
        title = g.graph["name"]

    edges = [list(d["kind"]) for _, _, d in g.edges(data=True)]
    edges = list(itertools.chain.from_iterable(edges))
    counts = pd.Series(Counter(edges))
    counts = pd.DataFrame(counts)
    counts.columns = ["counts"]

    if plot_type == "bar":
        fig = px.bar(counts, title=title)
    elif plot_type == "pie":
        fig = px.pie(counts, values="counts", names=counts.index, title=title)

    return fig


def graph_summary(
    G: nx.Graph,
    summary_statistics: List[str] = [
        "degree",
        "betweenness_centrality",
        "closeness_centrality",
        "eigenvector_centrality",
        "communicability_betweenness_centrality",
    ],
    custom_data: Optional[Union[pd.DataFrame, pd.Series]] = None,
    plot: bool = False,
) -> pd.DataFrame:
    """Returns a summary of the graph in a dataframe.

    :param G: NetworkX graph to get summary of.
    :type G: nx.Graph
    :param plot: Whether or not to plot the summary as a heatmap, defaults to ``False``.
    :type plot: bool
    :return: Dataframe of summary or plot.
    :rtype: pd.DataFrame
    """
    col_list = []
    col_names = []
    if "degree" in summary_statistics:
        degrees = pd.DataFrame(nx.degree(G))
        degrees.columns = ["node", "degree"]
        degrees.index = degrees["node"]
        degrees = degrees["degree"]
        col_list.append(degrees)
        col_names.append("degree")
    if "betweenness_centrality" in summary_statistics:
        betweenness = pd.Series(nx.betweenness_centrality(G))
        col_list.append(betweenness)
        col_names.append("betweenness_centrality")
    if "closeness_centrality" in summary_statistics:
        closeness = pd.Series(nx.closeness_centrality(G))
        col_list.append(closeness)
        col_names.append("closeness_centrality")
    if "eigenvector_centrality" in summary_statistics:
        eigenvector = pd.Series(nx.eigenvector_centrality_numpy(G))
        col_list.append(eigenvector)
        col_names.append("eigenvector_centrality")
    if "communicability_betweenness_centrality" in summary_statistics:
        communicability = pd.Series(
            nx.communicability_betweenness_centrality(G)
        )
        col_list.append(communicability)
        col_names.append("communicability_betweenness_centrality")

    df = pd.DataFrame(col_list).T
    df.columns = col_names
    df.index.name = "node"

    # Add custom data if provided
    if custom_data:
        df = pd.concat([df, custom_data], axis=1)

    if plot:
        return px.imshow(df.T)

    chain = [id.split(":")[0] for id in list(df.index)]
    residue_type = [id.split(":")[1] for id in list(df.index)]
    position = [id.split(":")[2] for id in list(df.index)]
    df["residue_type"] = residue_type
    df["position"] = position
    df["chain"] = chain

    return df


def plot_graph_metric_property_correlation(
    g: nx.Graph,
    summary_statistics: List[str] = [
        "degree",
        "betweenness_centrality",
        "closeness_centrality",
        "eigenvector_centrality",
        "communicability_betweenness_centrality",
    ],
    properties: List[str] = ["asa"],
    colour_by: Optional[str] = "residue_type",
    opacity: float = 0.2,
    diagonal_visible: bool = True,
    title: Optional[str] = None,
    height: int = 1000,
    width: int = 1000,
    font_size: int = 10,
) -> plotly.graph_objects.Figure:
    """Plots the correlation between graph metrics and properties.

    :param g: Protein graph to plot the correlation of.
    :type g: nx.Graph
    :param summary_statistics: List of graph metrics to employ in plot, defaults to
        ``["degree", "betweenness_centrality", "closeness_centrality", "eigenvector_centrality", "communicability_betweenness_centrality"]``.
    :type summary_statistics: List[str], optional
    :param properties: List of node properties to use in plot, defaults to ``["asa"]``.
    :type properties: List[str], optional
    :param colour_by: Controls colouring of points in plot. Options: ``"residue_type"``, ``"position"``, ``"chain"``, defaults to ``"residue_type"``.
    :type colour_by: Optional[str], optional
    :param opacity: Opacity of plot points, defaults to ``0.2``.
    :type opacity: float, optional
    :param diagonal_visible: Whether or not to show the diagonal plots, defaults to ``True``.
    :type diagonal_visible: bool, optional
    :param title: Title of plot, defaults to ``None``.
    :type title: Optional[str], optional
    :param height: Height of plot, defaults to ``1000``.
    :type height: int, optional
    :param width: Width of plot, defaults to ``1000``.
    :type width: int, optional
    :param font_size: Font size for plot text, defaults to ``10``.
    :type font_size: int, optional
    :return: Scatter plot matrix of graph metrics and protein properties.
    :rtype: plotly.graph_objects.Figure
    """
    if not title:
        title = (
            g.graph["name"] + " - Correlation of Graph Metrics with Properties"
        )
        if colour_by:
            title += f" (Coloured by {colour_by})"

    protein_properties = [
        pd.Series(nx.get_node_attributes(g, p), name=p) for p in properties
    ]
    protein_properties = pd.concat(protein_properties, axis=1)

    summary = graph_summary(g, summary_statistics=summary_statistics)
    summary = summary[
        summary_statistics + ["residue_type", "position", "chain"]
    ]
    dataf = pd.concat([summary, protein_properties], axis=1)

    fig = px.scatter_matrix(
        data_frame=dataf,
        # custom_data=dataf.index,
        dimensions=summary_statistics + properties,
        opacity=opacity,
        labels={
            col: col.replace("_", " ") for col in dataf.columns
        },  # remove underscore
        hover_name=dataf.index,
        color=colour_by,
        height=height,
        width=width,
        title=title,
    )
    fig.update_layout(font=dict(size=font_size))
    fig.update_traces(diagonal_visible=diagonal_visible)

    return fig



"""Functions to create protein meshes via pymol."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import importlib.util
import os
import time
from typing import List, NamedTuple, Optional, Tuple

from loguru import logger as log

from graphein.protein.config import ProteinMeshConfig
from graphein.utils.dependencies import (
    import_message,
    requires_external_dependencies,
    requires_python_libs,
)
from graphein.utils.pymol import MolViewer

try:
    from pytorch3d.structures import Meshes
except ImportError:
    message = import_message(
        submodule="graphein.protein.meshes",
        package="pytorch3d",
        conda_channel="pytorch3d",
        pip_install=True,
    )
    log.debug(message)


def check_for_pymol_installation():
    """Checks for presence of a pymol installation"""
    spec = importlib.util.find_spec("pymol")
    if spec is None:
        log.error(
            "Please install pymol: conda install -c schrodinger pymol or conda install -c tpeulen pymol-open-source"
        )


def configure_pymol_session(
    config: Optional[ProteinMeshConfig] = None,
):
    """
    Configures a PyMol session based on ``config.parse_pymol_commands``. Uses default parameters ``"-cKq"``.

    See: https://pymolwiki.org/index.php/Command_Line_Options

    :param config: :class:`~graphein.protein.config.ProteinMeshConfig` to use. Defaults to ``None`` which uses default config.
    :type config: graphein.protein.config.ProteinMeshConfig
    """
    pymol = MolViewer()
    pymol.delete("all")  # delete all objects from other sessions if necessary.

    # If no config is provided, use default
    if config is None:
        config = ProteinMeshConfig()

    # Start PyMol session
    pymol.start([config.pymol_command_line_options])


@requires_external_dependencies("pymol")
def get_obj_file(
    pdb_file: Optional[str] = None,
    pdb_code: Optional[str] = None,
    out_dir: Optional[str] = None,
    config: Optional[ProteinMeshConfig] = None,
) -> str:
    """
    Runs PyMol to compute surface/mesh for a given protein.

    :param pdb_file:  path to ``pdb_file`` to use. Defaults to ``None``.
    :type pdb_file: str, optional
    :param pdb_code: 4-letter pdb accession code. Defaults to ``None``.
    :type pdb_code: str, optional
    :param out_dir: path to output. Defaults to ``None``.
    :type out_dir: str, optional
    :param config: :class:`~graphein.protein.config.ProteinMeshConfig` containing pymol commands to run. Default is ``None`` (``"show surface"``).
    :type config: graphein.protein.config.ProteinMeshConfig
    :raises: ValueError if both or neither ``pdb_file`` or ``pdb_code`` are provided.
    :return: returns path to ``.obj`` file (str)
    :rtype: str
    """
    pymol = MolViewer()

    # Check inputs
    if not pdb_code and not pdb_file:
        raise ValueError("Please pass either a pdb_file or pdb_code argument")
    if pdb_code and pdb_file:
        raise ValueError(
            "Please pass either a pdb_file or pdb_code argument. Not both"
        )

    if out_dir is None:
        out_dir = "/tmp/"

    configure_pymol_session()

    # Load structure
    pymol.load(pdb_file) if pdb_file else pymol.fetch(pdb_code)
    # Create file_name
    file_name = (
        f"{pdb_file[:-3]}obj" if pdb_file else out_dir + pdb_code + ".obj"
    )

    if config is None:
        config = ProteinMeshConfig()

    commands = parse_pymol_commands(config)
    print(commands)
    run_pymol_commands(commands)

    # Save surface object as .obj
    pymol.do(f"save {file_name}")
    log.debug(f"Saved {file_name}")
    return file_name


def parse_pymol_commands(config: ProteinMeshConfig) -> List[str]:
    """
    Parses pymol commands from config. At the moment users can only supply a list of string commands.

    :param config: ProteinMeshConfig containing pymol commands to run in ``config.pymol_commands``.
    :type config: ProteinMeshConfig
    :return: list of pymol commands to run
    :rtype: List[str]
    """
    if config is None:
        config = ProteinMeshConfig()

    # Todo parsing of individual pymol mesh calculation parameters. There's a lot of them so this is not a priority now.
    if config.pymol_commands is not None:
        return config.pymol_commands


@requires_external_dependencies("pymol")
def run_pymol_commands(commands: List[str]) -> None:
    """
    Runs Pymol Commands.

    :param commands: List of commands to pass to PyMol.
    :type commands: List[str]
    """
    pymol = MolViewer()

    for c in commands:
        log.debug(c)
        pymol.do(c)


@requires_python_libs("pytorch3d")
def create_mesh(
    pdb_file: Optional[str] = None,
    pdb_code: Optional[str] = None,
    out_dir: Optional[str] = None,
    config: Optional[ProteinMeshConfig] = None,
) -> Tuple[torch.FloatTensor, NamedTuple, NamedTuple]:
    """
    Creates a ``PyTorch3D`` mesh from a ``pdb_file`` or ``pdb_code``.

    :param pdb_file: path to ``pdb_file``. Defaults to ``None``.
    :type pdb_file: str, optional
    :param pdb_code: 4-letter PDB accession code. Defaults to None.
    :type pdb_code: str, optional
    :param out_dir: output directory to store ``.obj`` file. Defaults to ``None``.
    :type out_dir: str, optional
    :param config:  :class:`~graphein.protein.config.ProteinMeshConfig` config to use. Defaults to default config in ``graphein.protein.config``.
    :type config: graphein.protein.config.ProteinMeshConfig
    :return: ``verts``, ``faces``, ``aux``.
    :rtype: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
    """
    from pytorch3d.io import load_obj

    if config is None:
        config = ProteinMeshConfig()

    obj_file = get_obj_file(
        pdb_code=pdb_code, pdb_file=pdb_file, out_dir=out_dir, config=config
    )
    # Wait for PyMol to finish
    while os.path.isfile(obj_file) is False:
        time.sleep(0.1)

    verts, faces, aux = load_obj(obj_file)
    return verts, faces, aux


@requires_python_libs("torch")
def normalize_and_center_mesh_vertices(
    verts: torch.FloatTensor,
) -> torch.FloatTensor:
    """
    We scale normalize and center the target mesh to fit in a sphere of radius 1 centered at ``(0,0,0)``.

    ``(scale, center)`` will be used to bring the predicted mesh to its original center and scale
    Note that normalizing the target mesh, speeds up the optimization but is not necessary!

    :param verts: Mesh vertices.
    :type verts: torch.FloatTensor
    :return: Normalized and centered vertices.
    :rtype: torch.FloatTensor
    """
    center = verts.mean()
    verts = verts - center
    scale = max(verts.abs().max(0)[0])
    verts = verts / scale
    return verts


@requires_python_libs("torch", "pytorch3d")
def convert_verts_and_face_to_mesh(
    verts: torch.FloatTensor, faces: NamedTuple
) -> Meshes:
    """
    Converts vertices and faces into a ``pytorch3d.structures`` Meshes object.

    :param verts: Vertices.
    :type verts: torch.FloatTensor
    :param faces: Faces.
    :type faces: NamedTuple
    :return: Meshes object.
    :rtype: pytorch3d.structures.Meshes
    """
    faces_idx = faces.verts_idx
    return Meshes(verts=[verts], faces=[faces_idx])


if __name__ == "__main__":
    from graphein.protein.visualisation import plot_pointcloud

    config = ProteinMeshConfig(
        pymol_commands=[
            "show surface",
            "set surface_solvent, on",
            "set solvent_radius, 10000",
        ]
    )
    verts, faces, aux = create_mesh(pdb_code="3eiy", config=config)

    trg_mesh = convert_verts_and_face_to_mesh(verts, faces)
    plot_pointcloud(trg_mesh)



"""Tests for graphein.ppi.graphs"""

import bioservices
import pytest

from graphein.ppi.config import PPIGraphConfig
from graphein.ppi.edges import add_biogrid_edges, add_string_edges
from graphein.ppi.features.node_features import add_sequence_to_nodes
from graphein.ppi.graphs import compute_ppi_graph

PROTEIN_LIST = [
    "CDC42",
    "CDK1",
    "KIF23",
    "PLK1",
    "RAC2",
    "RACGAP1",
    "RHOA",
    "RHOB",
]


@pytest.mark.skip("not a test")
def hgnc_available():
    try:
        a = bioservices.HGNC()
        return True
    except TypeError:
        return False


HGNC_AVAILABLE = hgnc_available()


# Test Graph Construction
@pytest.mark.skipif(not HGNC_AVAILABLE, reason="HGNC not available")
def test_construct_graph():
    config = PPIGraphConfig()

    g = compute_ppi_graph(
        protein_list=PROTEIN_LIST,
        edge_construction_funcs=[add_biogrid_edges, add_string_edges],
        node_annotation_funcs=[add_sequence_to_nodes],
        config=config,
    )

    # Check nodes and edges
    assert len(g.nodes()) == 8
    assert len(g.edges()) == 22

    # Check edge types are from string/biogrid
    # Check nodes are in our list
    for u, v, d in g.edges(data=True):
        assert d["kind"].issubset({"string", "biogrid"})
        assert u in PROTEIN_LIST
        assert v in PROTEIN_LIST

    # Check sequence is defined if UniProt ID found
    for _, d in g.nodes(data=True):
        assert d["protein_id"] in PROTEIN_LIST
        if d["uniprot_ids"] is not None:
            for id in d["uniprot_ids"]:
                assert d[f"sequence_{id}"] is not None



"""Functions for plotting protein graphs and meshes."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import re
from itertools import count
from typing import Dict, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import plotly.colors as co
import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns
from loguru import logger as log
from mpl_toolkits.mplot3d import Axes3D

from graphein.protein.subgraphs import extract_k_hop_subgraph
from graphein.utils.dependencies import import_message

try:
    from pytorch3d.ops import sample_points_from_meshes
except ImportError:
    message = import_message(
        submodule="graphein.protein.visualisation",
        package="pytorch3d",
        conda_channel="pytorch3d",
    )
    log.debug(message)

try:
    from mpl_chord_diagram import chord_diagram
except ImportError:
    import_message(
        submodule="graphein.protein.visualisation",
        package="mpl_chord_diagram",
        pip_install=True,
        extras=True,
    )
    log.debug(message)


def plot_pointcloud(mesh: Meshes, title: str = "") -> Axes3D:
    """
    Plots pytorch3d Meshes object as pointcloud.

    :param mesh: Meshes object to plot.
    :type mesh: pytorch3d.structures.meshes.Meshes
    :param title: Title of plot.
    :type title: str
    :return: returns Axes3D containing plot
    :rtype: Axes3D
    """
    # Sample points uniformly from the surface of the mesh.
    points = sample_points_from_meshes(mesh, 5000)
    x, y, z = points.clone().detach().cpu().squeeze().unbind(1)
    fig = plt.figure(figsize=(5, 5))
    ax = Axes3D(fig)
    ax.scatter3D(x, z, -y)
    ax.set_xlabel("x")
    ax.set_ylabel("z")
    ax.set_zlabel("y")
    ax.set_title(title)
    ax.view_init(190, 30)
    return ax


def colour_nodes(
    G: nx.Graph,
    colour_by: str,
    colour_map: matplotlib.colors.ListedColormap = plt.cm.plasma,
) -> List[Tuple[float, float, float, float]]:
    """
    Computes node colours based on ``"degree"``, ``"seq_position"`` or node
    attributes.

    :param G: Graph to compute node colours for
    :type G: nx.Graph
    :param colour_map:  Colourmap to use.
    :type colour_map: matplotlib.colors.ListedColormap
    :param colour_by: Manner in which to colour nodes. If not ``"degree"`` or
        ``"seq_position"``, this must correspond to a node feature.
    :type colour_by: str
    :return: List of node colours
    :rtype: List[Tuple[float, float, float, float]]
    """
    # get number of nodes
    n = G.number_of_nodes()

    # Define color range proportional to number of edges adjacent to a single
    # node
    if colour_by == "degree":
        # Get max number of edges connected to a single node
        edge_max = max(G.degree[i] for i in G.nodes())
        colors = [colour_map(G.degree[i] / (edge_max + 1)) for i in G.nodes()]
    elif colour_by == "seq_position":
        colors = [colour_map(i / n) for i in range(n)]
    elif colour_by == "chain":
        chains = G.graph["chain_ids"]
        chain_colours = dict(
            zip(chains, list(colour_map(1 / len(chains), 1, len(chains))))
        )
        colors = [chain_colours[d["chain_id"]] for _, d in G.nodes(data=True)]
    elif colour_by == "plddt":
        levels: List[str] = ["Very High", "Confident", "Low", "Very Low"]
        mapping = dict(zip(sorted(levels), count()))
        colors = []
        for _, d in G.nodes(data=True):
            if d["b_factor"] > 90:
                colors.append((27 / 256, 86 / 256, 206 / 256, 1))
            elif d["b_factor"] > 70:
                colors.append((126 / 256, 202 / 256, 239 / 256, 1))
            elif d["b_factor"] > 50:
                colors.append((250 / 256, 218 / 256, 77 / 256, 1))
            else:
                colors.append((239 / 256, 131 / 256, 83 / 256, 1))
        # colors = [colour_map(mapping[c] / len(levels)) for c in colors]
    else:
        node_types = set(nx.get_node_attributes(G, colour_by).values())
        mapping = dict(zip(sorted(node_types), count()))
        colors = [
            colour_map(mapping[d[colour_by]] / len(node_types))
            for n, d in G.nodes(data=True)
        ]

    return colors


def colour_edges(
    G: nx.Graph,
    colour_map: matplotlib.colors.ListedColormap,
    colour_by: Optional[str] = "kind",
    set_alpha: float = 1.0,
    return_as_rgba: bool = False,
) -> List[Tuple[float, float, float, float]] or List[str]:
    """
    Computes edge colours based on the kind of bond/interaction.

    :param G: nx.Graph protein structure graph to compute edge colours from.
    :type G: nx.Graph
    :param colour_map: Colourmap to use.
    :type colour_map: matplotlib.colors.ListedColormap
    :param colour_by: Edge attribute to colour by. Currently only ``"kind"`` is
        supported.
    :type colour_by: Optional[str]
    :param set_alpha: Sets a given alpha value between 0.0 and 1.0 for all the
        edge colours.
    :type set_alpha: float
    :param return_as_rgba: Returns a list of rgba strings instead of tuples.
    :type return_as_rgba: bool
    :return: List of edge colours.
    :rtype: List[Tuple[float, float, float, float]] or List[str]
    """
    if colour_by == "kind":
        edge_types = {
            frozenset(a) for a in nx.get_edge_attributes(G, "kind").values()
        }
        mapping = dict(zip(sorted(edge_types), count()))
        colors = [
            colour_map(
                mapping[frozenset(G.edges[i]["kind"])]
                / (len(edge_types) + 1)  # avoid division by zero error
            )
            for i in G.edges()
        ]
    elif colour_by is None:
        colors = [(0.0, 0.0, 0.0, 1.0) for _ in G.edges()]
    else:
        edge_types = set(nx.get_edge_attributes(G, colour_by).values())
        mapping = dict(zip(sorted(edge_types), count()))
        colors = [
            colour_map(mapping[d[colour_by]] / (len(edge_types) + 1))
            for _, _, d in G.edges(data=True)
        ]

    assert (
        0.0 <= set_alpha <= 1.0
    ), f"Alpha value {set_alpha} must be between 0.0 and 1.0"
    colors = [c[:3] + (set_alpha,) for c in colors]
    if return_as_rgba:
        return [
            f"rgba{tuple(list(co.convert_to_RGB_255(c[:3])) + [c[3]])}"
            for c in colors
        ]
    return colors


def plotly_protein_structure_graph(
    G: nx.Graph,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (620, 650),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 1.0,
    node_size_feature: str = "degree",
    label_node_ids: bool = True,
    node_colour_map=plt.cm.plasma,
    edge_color_map=plt.cm.plasma,
    colour_nodes_by: str = "degree",
    colour_edges_by: Optional[str] = "kind",
) -> go.Figure:
    """
    Plots protein structure graph using plotly.

    :param G:  nx.Graph Protein Structure graph to plot
    :type G: nx.Graph
    :param plot_title: Title of plot, defaults to ``None``.
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to ``(620, 650)``.
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to ``0.7``.
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size. Defaults to ``20.0``.
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes
        reflect degree. Defaults to ``1.0``.
    :type node_size_multiplier: float
    :param node_size_feature: Which feature to scale the node size by. Defaults
        to ``degree``.
    :type node_size_feature: str
    :param label_node_ids: bool indicating whether or not to plot ``node_id``
        labels. Defaults to ``True``.
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes. Defaults to
        ``plt.cm.plasma``.
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges. Defaults to
        ``plt.cm.plasma``.
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. ``"degree"``,
        ``"seq_position"`` or a node feature.
    :type colour_nodes_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only
        ``"kind"`` or ``None`` are supported.
    :type colour_edges_by: Optional[str]
    :returns: Plotly Graph Objects plot
    :rtype: go.Figure
    """

    # Get Node Attributes
    pos = nx.get_node_attributes(G, "coords")

    # Get node colours
    node_colors = colour_nodes(
        G, colour_map=node_colour_map, colour_by=colour_nodes_by
    )
    edge_colors = colour_edges(
        G, colour_map=edge_color_map, colour_by=colour_edges_by
    )

    # Get node size
    def node_scale_by(G: nx.Graph, feature: str):
        if feature == "degree":
            return lambda k: node_size_min + node_size_multiplier * G.degree[k]
        elif feature == "rsa":
            return (
                lambda k: node_size_min
                + node_size_multiplier * G.nodes(data=True)[k]["rsa"]
            )

        # Meiler embedding dimension
        p = re.compile("meiler-([1-7])")
        dim = p.search(feature)[1]
        if dim:
            return lambda k: node_size_min + node_size_multiplier * max(
                0, G.nodes(data=True)[k]["meiler"][f"dim_{dim}"]
            )  # Meiler values may be negative
        else:
            raise ValueError(f"Cannot size nodes by feature '{feature}'")

    get_node_size = node_scale_by(G, node_size_feature)

    # 3D network plot
    x_nodes = []
    y_nodes = []
    z_nodes = []
    node_sizes = []
    node_labels = []

    # Loop on the pos dictionary to extract the x,y,z coordinates of each node
    for i, (key, value) in enumerate(pos.items()):
        x_nodes.append(value[0])
        y_nodes.append(value[1])
        z_nodes.append(value[2])
        node_sizes.append(get_node_size(key))

        if label_node_ids:
            node_labels.append(list(G.nodes())[i])

    nodes = go.Scatter3d(
        x=x_nodes,
        y=y_nodes,
        z=z_nodes,
        mode="markers",
        marker={
            "symbol": "circle",
            "color": node_colors,
            "size": node_sizes,
            "opacity": node_alpha,
        },
        text=list(G.nodes()),
        hoverinfo="text+x+y+z",
    )

    # Loop on the list of edges to get the x,y,z, coordinates of the connected
    # nodes. Those two points are the extrema of the line to be plotted
    x_edges = []
    y_edges = []
    z_edges = []

    for node_a, node_b in G.edges(data=False):
        x_edges.extend([pos[node_a][0], pos[node_b][0], None])
        y_edges.extend([pos[node_a][1], pos[node_b][1], None])
        z_edges.extend([pos[node_a][2], pos[node_b][2], None])

    axis = dict(
        showbackground=False,
        showline=False,
        zeroline=False,
        showgrid=False,
        showticklabels=False,
        title="",
    )

    repeated_edge_colours = []
    for (
        edge_col
    ) in (
        edge_colors
    ):  # Repeat as each line segment is ({x,y,z}_start, {x,y,z}_end, None)
        repeated_edge_colours.extend((edge_col, edge_col, edge_col))

    edge_colors = repeated_edge_colours

    edge_text = [
        " / ".join(list(edge_type))
        for edge_type in nx.get_edge_attributes(G, "kind").values()
    ]
    edge_text = np.repeat(
        edge_text, 3
    )  # Repeat as each line segment is ({x,y,z}_start, {x,y,z}_end, None)

    edges = go.Scatter3d(
        x=x_edges,
        y=y_edges,
        z=z_edges,
        mode="lines",
        line={"color": edge_colors, "width": 10},
        text=edge_text,
        hoverinfo="text",
    )

    return go.Figure(
        data=[nodes, edges],
        layout=go.Layout(
            title=plot_title,
            width=figsize[0],
            height=figsize[1],
            showlegend=False,
            scene=dict(
                xaxis=dict(axis),
                yaxis=dict(axis),
                zaxis=dict(axis),
            ),
            margin=dict(t=100),
        ),
    )


def plot_protein_structure_graph(
    G: nx.Graph,
    angle: int = 30,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (10, 7),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 1.0,
    label_node_ids: bool = True,
    node_colour_map=plt.cm.plasma,
    edge_color_map=plt.cm.plasma,
    colour_nodes_by: str = "degree",
    colour_edges_by: str = "kind",
    edge_alpha: float = 0.5,
    plot_style: str = "ggplot",
    out_path: Optional[str] = None,
    out_format: str = ".png",
) -> Axes3D:
    """
    Plots protein structure graph in ``Axes3D``.

    :param G:  nx.Graph Protein Structure graph to plot.
    :type G: nx.Graph
    :param angle:  View angle. Defaults to ``30``.
    :type angle: int
    :param plot_title: Title of plot. Defaults to ``None``.
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to ``(10, 7)``.
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to ``0.7``.
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size, defaults to ``20.0``.
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes
        reflect degree. Defaults to ``1.0``.
    :type node_size_multiplier: float
    :param label_node_ids: bool indicating whether or not to plot ``node_id``
        labels. Defaults to ``True``.
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes. Defaults to
        ``plt.cm.plasma``.
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges. Defaults to
        ``plt.cm.plasma``.
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. ``"degree"``,
        ``"seq_position"`` or a node feature.
    :type colour_nodes_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only
        ``"kind"`` is supported.
    :type colour_edges_by: str
    :param edge_alpha: Controls edge transparency. Defaults to ``0.5``.
    :type edge_alpha: float
    :param plot_style: matplotlib style sheet to use. Defaults to ``"ggplot"``.
    :type plot_style: str
    :param out_path: If not none, writes plot to this location. Defaults to
        ``None`` (does not save).
    :type out_path: str, optional
    :param out_format: File format to use for plot
    :type out_format: str
    :return: matplotlib Axes3D object.
    :rtype: Axes3D
    """

    # Get Node Attributes
    pos = nx.get_node_attributes(G, "coords")

    # Get node colours
    node_colors = colour_nodes(
        G, colour_map=node_colour_map, colour_by=colour_nodes_by
    )
    edge_colors = colour_edges(
        G, colour_map=edge_color_map, colour_by=colour_edges_by
    )

    # 3D network plot
    with plt.style.context(plot_style):
        fig = plt.figure(figsize=figsize)
        ax = Axes3D(fig)
        fig.add_axes(ax)

        # Loop on the pos dictionary to extract the x,y,z coordinates of each
        # node
        for i, (key, value) in enumerate(pos.items()):
            xi = value[0]
            yi = value[1]
            zi = value[2]

            # Scatter plot
            ax.scatter(
                xi,
                yi,
                zi,
                color=node_colors[i],
                s=node_size_min + node_size_multiplier * G.degree[key],
                edgecolors="k",
                alpha=node_alpha,
            )
            if label_node_ids:
                label = list(G.nodes())[i]
                ax.text(xi, yi, zi, label)

        # Loop on the list of edges to get the x,y,z, coordinates of the
        # connected nodes. Those two points are the extrema of the line to be
        # plotted
        for i, j in enumerate(G.edges()):
            x = np.array((pos[j[0]][0], pos[j[1]][0]))
            y = np.array((pos[j[0]][1], pos[j[1]][1]))
            z = np.array((pos[j[0]][2], pos[j[1]][2]))

            # Plot the connecting lines
            ax.plot(x, y, z, c=edge_colors[i], alpha=edge_alpha)

    # Set title
    ax.set_title(plot_title)
    # Set the initial view
    ax.view_init(30, angle)
    # Hide the axes
    ax.set_axis_off()
    if out_path is not None:
        plt.savefig(out_path + str(angle).zfill(3) + out_format)
        plt.close("all")

    return ax


def add_vector_to_plot(
    g: nx.Graph,
    fig,
    vector: str = "sidechain_vector",
    scale: float = 5,
    colour: str = "red",
    width: int = 10,
) -> go.Figure:
    """Adds representations of vector features to the protein graph.

    Requires all nodes have a vector feature (1 x 3 array).

    :param g: Protein graph containing vector features
    :type g: nx.Graph
    :param fig: 3D plotly figure to add vectors to.
    :type fig: go.Figure
    :param vector: Name of node vector feature to add, defaults to
        ``"sidechain_vector"``.
    :type vector: str, optional
    :param scale: How much to scale the vectors by, defaults to 5
    :type scale: float, optional
    :param colour: Colours for vectors, defaults to ``"red"``.
    :type colour: str, optional
    :return: 3D Plotly plot with vectors added.
    :rtype: go.Figure
    """
    # Compute line segment positions
    x_edges = []
    y_edges = []
    z_edges = []
    edge_text = []
    for _, d in g.nodes(data=True):
        x_edges.extend(
            [d["coords"][0], d["coords"][0] + d[vector][0] * scale, None]
        )
        y_edges.extend(
            [d["coords"][1], d["coords"][1] + d[vector][1] * scale, None]
        )
        z_edges.extend(
            [d["coords"][2], d["coords"][2] + d[vector][2] * scale, None]
        )
        edge_text.extend([None, f"{vector}", None])

    edge_trace = go.Scatter3d(
        x=x_edges,
        y=y_edges,
        z=z_edges,
        mode="lines",
        line={"color": colour, "width": width},
        text=3 * [f"{vector}" for _ in range(len(g))],
        hoverinfo="text",
    )
    # Compute cone positions.
    arrow_tip_ratio = 0.1
    arrow_starting_ratio = 0.98
    x = []
    y = []
    z = []
    u = []
    v = []
    w = []
    for _, d in g.nodes(data=True):
        x.extend(
            [d["coords"][0] + d[vector][0] * scale * arrow_starting_ratio]
        )
        y.extend(
            [d["coords"][1] + d[vector][1] * scale * arrow_starting_ratio]
        )
        z.extend(
            [d["coords"][2] + d[vector][2] * scale * arrow_starting_ratio]
        )
        u.extend([d[vector][0]])  # * arrow_tip_ratio])
        v.extend([d[vector][1]])  # * arrow_tip_ratio])
        w.extend([d[vector][2]])  # * arrow_tip_ratio])

    if colour == "red":
        colour = [[0, "rgb(255,0,0)"], [1, "rgb(255,0,0)"]]
    elif colour == "blue":
        colour = [[0, "rgb(0,0,255)"], [1, "rgb(0,0,255)"]]
    elif colour == "green":
        colour = [[0, "rgb(0,255,0)"], [1, "rgb(0,255,0)"]]

    cone_trace = go.Cone(
        x=x,
        y=y,
        z=z,
        u=u,
        v=v,
        w=w,
        text=[f"{vector}" for _ in range(len(g.nodes()))],
        hoverinfo="u+v+w+text",
        colorscale=colour,
        showlegend=False,
        showscale=False,
        sizemode="absolute",
    )
    fig.add_trace(edge_trace)
    fig.add_trace(cone_trace)
    return fig


def plot_distance_matrix(
    g: Optional[nx.Graph],
    dist_mat: Optional[np.ndarray] = None,
    use_plotly: bool = True,
    title: Optional[str] = None,
    show_residue_labels: bool = True,
) -> go.Figure:
    """Plots a distance matrix of the graph.

    :param g: NetworkX graph containing a distance matrix as a graph attribute
        (``g.graph['dist_mat']``).
    :type g: nx.Graph, optional
    :param dist_mat: Distance matrix to plot. If not provided, the distance
        matrix is taken from the graph. Defaults to ``None``.
    :type dist_mat: np.ndarray, optional
    :param use_plotly: Whether to use ``plotly`` or ``seaborn`` for plotting.
        Defaults to ``True``.
    :type use_plotly: bool
    :param title: Title of the plot.Defaults to ``None``.
    :type title: str, optional
    :show_residue_labels: Whether to show residue labels on the plot. Defaults
        to ``True``.
    :type show_residue_labels: bool
    :raises: ValueError if neither a graph ``g`` or a ``dist_mat`` are provided.
    :return: Plotly figure.
    :rtype: px.Figure
    """
    if g is None and dist_mat is None:
        raise ValueError("Must provide either a graph or a distance matrix.")

    if dist_mat is None:
        dist_mat = g.graph["dist_mat"]  # type: ignore
    if g is not None:
        x_range = list(g.nodes)
        y_range = list(g.nodes)
        if not title:
            title = g.graph["name"] + " - Distance Matrix"
    else:
        x_range = list(range(dist_mat.shape[0]))  # type: ignore
        y_range = list(range(dist_mat.shape[1]))  # type: ignore
        if not title:
            title = "Distance matrix"

    if use_plotly:
        return px.imshow(
            dist_mat,
            x=x_range,
            y=y_range,
            labels=dict(color="Distance"),
            title=title,
        )
    tick_labels = x_range if show_residue_labels else []
    return sns.heatmap(
        dist_mat, xticklabels=tick_labels, yticklabels=tick_labels
    ).set(title=title)


def plot_distance_landscape(
    g: Optional[nx.Graph] = None,
    dist_mat: Optional[np.ndarray] = None,
    add_contour: bool = True,
    title: Optional[str] = None,
    width: int = 500,
    height: int = 500,
    autosize: bool = False,
) -> go.Figure:
    """Plots a distance landscape of the graph.

    :param g: Graph to plot (must contain a distance matrix in
        ``g.graph["dist_mat"]``).
    :type g: nx.Graph
    :param add_contour: Whether or not to show the contour, defaults to
        ``True``.
    :type add_contour: bool, optional
    :param width: Plot width, defaults to ``500``.
    :type width: int, optional
    :param height: Plot height, defaults to ``500``.
    :type height: int, optional
    :param autosize: Whether or not to autosize the plot, defaults to ``False``.
    :type autosize: bool, optional
    :return: Plotly figure of distance landscape.
    :rtype: go.Figure
    """
    if g:
        dist_mat = g.graph["dist_mat"]
        if not title:
            title = g.graph["name"] + " - Distance Landscape"
        tick_labels = list(g.nodes)
    else:
        if not title:
            title = "Distance landscape"
        tick_labels = list(range(dist_mat.shape[0]))

    fig = go.Figure(data=[go.Surface(z=dist_mat)])

    if add_contour:
        fig.update_traces(
            contours_z=dict(
                show=True,
                usecolormap=True,
                highlightcolor="limegreen",
                project_z=True,
            )
        )

    fig.update_layout(
        title=title,
        autosize=autosize,
        width=width,
        height=height,
        scene=dict(
            zaxis_title="Distance",
            xaxis=dict(
                ticktext=tick_labels,
                tickvals=list(range(len(tick_labels))),
                nticks=10,
                showticklabels=False,
            ),
            yaxis=dict(
                ticktext=tick_labels,
                tickvals=list(range(len(tick_labels))),
                nticks=10,
                showticklabels=False,
            ),
        ),
    )

    return fig


def asteroid_plot(
    g: nx.Graph,
    node_id: str,
    k: int = 2,
    colour_nodes_by: str = "shell",  # residue_name
    colour_edges_by: str = "kind",
    edge_colour_map: plt.cm.Colormap = plt.cm.plasma,
    edge_alpha: float = 1.0,
    show_labels: bool = True,
    title: Optional[str] = None,
    width: int = 600,
    height: int = 500,
    use_plotly: bool = True,
    show_edges: bool = False,
    show_legend: bool = True,
    node_size_multiplier: float = 10.0,
) -> Union[plotly.graph_objects.Figure, matplotlib.figure.Figure]:
    # sourcery skip: remove-unnecessary-else, swap-if-else-branches
    """Plots a k-hop subgraph around a node as concentric shells.

    Radius of each point is proportional to the degree of the node
        (modified by node_size_multiplier).

    :param g: NetworkX graph to plot.
    :type g: nx.Graph
    :param node_id: Node to centre the plot around.
    :type node_id: str
    :param k: Number of hops to plot. Defaults to ``2``.
    :type k: int
    :param colour_nodes_by: Colour the nodes by this attribute. Currently only
        ``"shell"`` is supported.
    :type colour_nodes_by: str
    :param colour_edges_by: Colour the edges by this attribute. Currently only
        ``"kind"`` is supported.
    :type colour_edges_by: str
    :param edge_colour_map: Colour map for edges. Defaults to ``plt.cm.plasma``.
    :type edge_colour_map: plt.cm.Colormap
    :param edge_alpha: Sets a given alpha value between 0.0 and 1.0 for all the
        edge colours.
    :type edge_alpha: float
    :param title: Title of the plot. Defaults to ``None``.
    :type title: str
    :param width: Width of the plot. Defaults to ``600``.
    :type width: int
    :param height: Height of the plot. Defaults to ``500``.
    :type height: int
    :param use_plotly: Use plotly to render the graph. Defaults to ``True``.
    :type use_plotly: bool
    :param show_edges: Whether to show edges in the plot. Defaults to ``False``.
    :type show_edges: bool
    :param show_legend: Whether to show the legend of the edges. Defaults to
        `True``.
    :type show_legend: bool
    :param node_size_multiplier: Multiplier for the size of the nodes. Defaults
        to ``10.0``.
    :type node_size_multiplier: float
    :returns: Plotly figure or matplotlib figure.
    :rtpye: Union[plotly.graph_objects.Figure, matplotlib.figure.Figure]
    """
    assert node_id in g.nodes(), f"Node {node_id} not in graph"

    nodes: Dict[int, List[str]] = {0: [node_id]}
    node_list: List[str] = [node_id]
    # Iterate over the number of hops and extract nodes in each shell
    for i in range(1, k):
        subgraph = extract_k_hop_subgraph(g, node_id, k=i)
        candidate_nodes = subgraph.nodes()
        # Check we've not already found nodes in the previous shells
        nodes[i] = [n for n in candidate_nodes if n not in node_list]
        node_list += candidate_nodes
    shells = [nodes[i] for i in range(k)]
    log.debug(f"Plotting shells: {shells}")

    if use_plotly:
        # Get shell layout and set as node attributes.
        pos = nx.shell_layout(subgraph, shells)
        nx.set_node_attributes(subgraph, pos, "pos")

        if show_edges:
            edge_colors = colour_edges(
                subgraph,
                colour_map=edge_colour_map,
                colour_by=colour_edges_by,
                set_alpha=edge_alpha,
                return_as_rgba=True,
            )
            show_legend_bools = [
                x not in edge_colors[:i] for i, x in enumerate(edge_colors)
            ]
            edge_trace = []
            for i, (u, v) in enumerate(subgraph.edges()):
                x0, y0 = subgraph.nodes[u]["pos"]
                x1, y1 = subgraph.nodes[v]["pos"]
                bond_kind = " / ".join(list(subgraph[u][v]["kind"]))
                tr = go.Scatter(
                    x=(x0, x1),
                    y=(y0, y1),
                    mode="lines",
                    line=dict(width=1, color=edge_colors[i]),
                    hoverinfo="text",
                    text=[bond_kind],
                    name=bond_kind,
                    legendgroup=bond_kind,
                    showlegend=show_legend_bools[i],
                )
                edge_trace.append(tr)

        node_x: List[str] = []
        node_y: List[str] = []
        for node in subgraph.nodes():
            x, y = subgraph.nodes[node]["pos"]
            node_x.append(x)
            node_y.append(y)

        degrees = [
            subgraph.degree(n) * node_size_multiplier for n in subgraph.nodes()
        ]

        if colour_nodes_by == "shell":
            node_colours = []
            for n in subgraph.nodes():
                for k, v in nodes.items():
                    if n in v:
                        node_colours.append(k)
        else:
            raise NotImplementedError(
                f"Colour by {colour_nodes_by} not implemented."
            )
            # TODO colour by AA type
        node_trace = go.Scatter(
            x=node_x,
            y=node_y,
            text=list(subgraph.nodes()),
            mode="markers+text" if show_labels else "markers",
            hoverinfo="text",
            textposition="bottom center",
            showlegend=False,
            marker=dict(
                colorscale="YlGnBu",
                reversescale=True,
                color=node_colours,
                size=degrees,
                colorbar=dict(
                    thickness=15,
                    title="Shell",
                    tickvals=list(range(k)),
                    xanchor="left",
                ),
                line_width=2,
            ),
        )

        data = edge_trace + [node_trace] if show_edges else [node_trace]
        return go.Figure(
            data=data,
            layout=go.Layout(
                title=title or f'Asteroid Plot - {g.graph["name"]}',
                width=width,
                height=height,
                legend=dict(yanchor="top", y=1, xanchor="left", x=1.10),
                showlegend=show_legend,
                hovermode="closest",
                margin=dict(b=20, l=5, r=5, t=40),
                xaxis=dict(
                    showgrid=False, zeroline=False, showticklabels=False
                ),
                yaxis=dict(
                    showgrid=False, zeroline=False, showticklabels=False
                ),
            ),
        )
    else:
        nx.draw_shell(subgraph, nlist=shells, with_labels=show_labels)


def plot_chord_diagram(
    g: nx.Graph,
    show_names: bool = True,
    order: Optional[List] = None,
    width: float = 0.1,
    pad: float = 2.0,
    gap: float = 0.03,
    chordwidth: float = 0.7,
    ax=None,
    colors=None,
    cmap=None,
    alpha=0.7,
    use_gradient: bool = False,
    chord_colors=None,
    show: bool = False,
    **kwargs,
):
    """
    Plot a chord diagram.


    Based on Tanguy Fardet's implementation:
    https://github.com/tfardet/mpl_chord_diagram

    :param g: NetworkX graph to plot
        Flux data, mat[i, j] is the flux from i to j (adjacency matrix)
    :type g: nx.Graph
    :param show_names: Whether to show the names of the nodes
    :type show_names: bool
    :param order: list, optional (default: order of the matrix entries)
        Order in which the arcs should be placed around the trigonometric
        circle.
    :param width: float, optional (default: 0.1)
        Width/thickness of the ideogram arc.
    :type width: float
    :param pad: float, optional (default: 2)
        Distance between two neighboring ideogram arcs. Unit: degree.
    :type pad: float
    :param gap: float, optional (default: 0)
        Distance between the arc and the beginning of the cord.
    :type gap: float
    :param chordwidth: float, optional (default: 0.7)
        Position of the control points for the chords, controlling their shape.
    :param ax: matplotlib axis, optional (default: new axis)
        Matplotlib axis where the plot should be drawn.
    :param colors: list, optional (default: from `cmap`)
        List of user defined colors or floats.
    :param cmap: str or colormap object (default: viridis)
        Colormap that will be used to color the arcs and chords by default.
        See `chord_colors` to use different colors for chords.
    :param alpha: float in [0, 1], optional (default: 0.7)
        Opacity of the chord diagram.
    :param use_gradient: bool, optional (default: False)
        Whether a gradient should be use so that chord extremities have the
        same color as the arc they belong to.
    :type use_gradient: bool
    :param chord_colors: str, or list of colors, optional (default: None)
        Specify color(s) to fill the chords differently from the arcs.
        When the keyword is not used, chord colors default to the colomap given
        by `colors`.
        Possible values for `chord_colors` are:

        * a single color (do not use an RGB tuple, use hex format instead),
        e.g. "red" or "#ff0000"; all chords will have this color
        * a list of colors, e.g. ``["red", "green", "blue"]``, one per node
        (in this case, RGB tuples are accepted as entries to the list).
        Each chord will get its color from its associated source node, or
        from both nodes if `use_gradient` is True.
    :param show: bool, optional (default: False)
        Whether the plot should be displayed immediately via an automatic call
        to ``plt.show()``.
    :param kwargs: keyword arguments
        Available kwargs are:

        ================  ==================  ===============================
            Name               Type           Purpose and possible values
        ================  ==================  ===============================
        fontcolor         str or list         Color of the names
        fontsize          int                 Size of the font for names
        rotate_names      (list of) bool(s)   Rotate names by 90°
        sort              str                 Either "size" or "distance"
        zero_entry_size   float               Size of zero-weight reciprocal
        ================  ==================  ===============================
    :type kwargs: Dict[str, Any]
    """
    mat = nx.adjacency_matrix(g).todense()
    names = list(g.nodes)
    if show_names:
        if g.graph["node_type"] == "chain":
            names = [f"Chain {n}" for n in names]
        elif g.graph["node_type"] != "secondary_structure":
            raise ValueError()
    else:
        names = None

    a = chord_diagram(
        mat,
        names=names,
        order=order,
        width=width,
        pad=pad,
        gap=gap,
        chordwidth=chordwidth,
        ax=ax,
        colors=colors,
        cmap=cmap,
        alpha=alpha,
        use_gradient=use_gradient,
        chord_colors=chord_colors,
        show=show,
        **kwargs,
    )



from functools import partial

import pytest

from graphein.grn.config import GRNGraphConfig
from graphein.grn.edges import add_regnetwork_edges, add_trrust_edges
from graphein.grn.features.node_features import add_sequence_to_nodes
from graphein.grn.graphs import compute_grn_graph
from graphein.utils.utils import ping

GENE_LIST = ["AATF", "MYC", "USF1", "SP1", "TP53", "DUSP1"]


@pytest.mark.skipif(
    not ping("regnetworkweb.org"),
    reason="RegNetwork Web is intermittently down",
)
def test_construct_graph():
    config = GRNGraphConfig()

    g = compute_grn_graph(
        gene_list=GENE_LIST,
        edge_construction_funcs=[
            partial(
                add_trrust_edges,
                trrust_filtering_funcs=config.trrust_config.filtering_functions,
            ),
            partial(
                add_regnetwork_edges,
                regnetwork_filtering_funcs=config.regnetwork_config.filtering_functions,
            ),
        ],
        node_annotation_funcs=[add_sequence_to_nodes],  # , molecular_weight],
        edge_annotation_funcs=[],
    )
    print(g.edges(data=True))


def test_grn_config():
    config = GRNGraphConfig()
    assert config is not None


if __name__ == "__main__":
    test_construct_graph()



"""Utilities for parsing proteins into and out of tensors."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import collections
import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from biopandas.pdb import PandasPdb
from loguru import logger as log

from graphein.utils.dependencies import import_message

from ..graphs import (
    deprotonate_structure,
    filter_hetatms,
    read_pdb_to_dataframe,
    remove_insertions,
    select_chains,
    sort_dataframe,
)
from ..resi_atoms import (
    ATOM_NUMBERING,
    ELEMENT_SYMBOL_MAP,
    PROTEIN_ATOMS,
    STANDARD_AMINO_ACID_MAPPING_1_TO_3,
    STANDARD_AMINO_ACIDS,
)
from .representation import get_full_atom_coords
from .sequence import (
    get_residue_id,
    get_sequence,
    infer_residue_types,
    residue_type_tensor,
)
from .types import AtomTensor

try:
    from torch_geometric.data import Data
except ImportError:
    message = import_message(
        submodule="graphein.protein.tensor.io",
        package="torch_geometric",
        conda_channel="pyg",
        pip_install=True,
    )
    log.warning(message)

try:
    import torch
except ImportError:
    message = import_message(
        submodule="graphein.protein.tensor.io",
        package="torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.warning(message)


def get_protein_length(df: pd.DataFrame, insertions: bool = True) -> int:
    """Return the number of unique amino acids in the protein.

    Note for future development: this function may return incorrect results
    on raw PDB files containing insertions/altlocs. This should not be the case
    for proteins processed by the Graphein pre-processing workflow. Caveat
    emptor.

    :param df: Protein DataFrame to get length of.
    :type df: pd.DataFrame
    :param insertions: Whether or not to include insertions in the length.
        Defaults to ``True``.
    :type insertions: bool
    :return: Number of unique residues.
    :rtype: int
    """
    # Get unique residues:
    if "residue_id" not in df.columns:
        df["residue_id"] = (
            df["chain_id"]
            + ":"
            + df["residue_name"]
            + ":"
            + df["residue_number"].astype(str)
        )
        if insertions:
            df["residue_id"] = df.residue_id + ":" + df.insertion
    ids = df["residue_id"].values

    return len(set(ids))


def protein_to_pyg(
    path: Optional[Union[str, os.PathLike]] = None,
    pdb_code: Optional[str] = None,
    uniprot_id: Optional[str] = None,
    df: Optional[pd.DataFrame] = None,
    chain_selection: Union[str, List[str]] = "all",
    deprotonate: bool = True,
    keep_insertions: bool = True,
    keep_hets: List[str] = [],
    model_index: int = 1,
    atom_types: List[str] = PROTEIN_ATOMS,
    remove_nonstandard: bool = True,
    store_het: bool = False,
    store_bfactor: bool = False,
    fill_value_coords: float = 1e-5,
) -> Data:
    """
    Parses a protein (from either: a PDB code, PDB file or a UniProt ID
    (via AF2 database) to a PyTorch Geometric ``Data`` object.


    .. code-block:: python

        import graphein.protein.tensor as gpt

        # From PDB code
        gpt.io.protein_to_pyg(pdb_code="3eiy")

        # From PDB Path
        gpt.io.protein_to_pyg(path="3eiy.pdb")

        # From MMTF Path
        gpt.io.protein_to_pyg(path="3eiy.mmtf")

        # From UniProt ID
        gpt.io.protein_to_pyg(uniprot_id="Q5VSL9")


    :param path: Path to PDB or MMTF file. Default is ``None``.
    :type path: Union[str, os.PathLike]
    :param pdb_code: PDB accesion code. Default is ``None``.
    :type pdb_code: str
    :param uniprot_id: UniProt ID. Default is ``None``.
    :type uniprot_id: str
    :param chain_selection: Selection of chains to include (e.g.
        ``["A", "C", "AB"]``) or ``"all"``. Default is ``"all"``.
    :type chain_selection: Union[str, List[str]]
    :param deprotonate: Whether or not to remove Hydrogens. Default is ``True``.
    :type deprotonate: bool
    :param keep_insertions: Whether or not to keep insertions. Default is
        ``True``.
    :type keep_insertions: bool
    :param keep_hets: List of heteroatoms to include. E.g. ``["HOH"]``.
    :type keep_hets: List[str]
    :param model_index: Index of model in models containing multiple structures.
    :type model_index: int
    :param atom_types: List of atom types to select. Default is:
        :const:`graphein.protein.resi_atoms.PROTEIN_ATOMS`
    :type atom_types: List[str]
    :param remove_nonstandard: Whether or not to remove non-standard residues.
        Default is ``True``.
    :type remove_nonstandard: bool
    :param store_het: Whether or not to store heteroatoms in the ``Data``
        object. Default is ``False``.
    :type store_het: bool
    :param store_bfactor: Whether or not to store bfactors in the ``Data``
        object. Default is ``False.
    :type store_bfactor: bool
    :param fill_value_coords: Fill value to use for positions in atom37
        representation that are not filled. Defaults to 1e-5
    :type fill_value_coords: float
    :returns: ``Data`` object with attributes: ``x`` (AtomTensor), ``residues``
        (list of 3-letter residue codes), id (ID of protein), residue_id (E.g.
        ``"A:SER:1"``), residue_type (torch.Tensor), ``chains`` (torch.Tensor).
    :rtype: torch_geometric.data.Data
    """

    # Get ID
    if path is not None:
        id = (
            os.path.splitext(path)[0].split("/")[-1]
            + "_"
            + "".join(chain_selection)
            if chain_selection != "all"
            else os.path.splitext(path)[0].split("/")[-1]
        )
    elif pdb_code is not None:
        id = (
            pdb_code + "_" + "".join(chain_selection)
            if chain_selection != "all"
            else pdb_code
        )
    elif uniprot_id is not None:
        id = (
            uniprot_id + "_" + "".join(chain_selection)
            if chain_selection != "all"
            else uniprot_id
        )
    else:
        id = None

    if df is None:
        df = read_pdb_to_dataframe(
            path=path,
            pdb_code=pdb_code,
            uniprot_id=uniprot_id,
            model_index=model_index,
        )
    if chain_selection != "all":
        if isinstance(chain_selection, str):
            chain_selection = [chain_selection]
        df = select_chains(df, chain_selection)

    if deprotonate:
        df = deprotonate_structure(df)
    if not keep_insertions:
        df = remove_insertions(df)
    # Remove hetatms
    hets = filter_hetatms(df, keep_hets=keep_hets)

    if store_het:
        hetatms = df.loc[df.record_name == "HETATM"]
        all_hets = list(set(hetatms.residue_name))
        het_data = collections.defaultdict(dict)
        for het in all_hets:
            het_data[het]["coords"] = torch.tensor(
                hetatms.loc[hetatms.residue_name == het][
                    ["x_coord", "y_coord", "z_coord"]
                ].values
            )
            het_data[het]["atoms"] = hetatms.loc[hetatms.residue_name == het][
                "atom_name"
            ].values
            het_data[het]["residue_number"] = torch.tensor(
                hetatms.loc[hetatms.residue_name == het][
                    "residue_number"
                ].values
            )
            het_data[het]["element_symbol"] = hetatms.loc[
                hetatms.residue_name == het
            ]["element_symbol"].values

    df = df.loc[df.record_name == "ATOM"]
    if remove_nonstandard:
        df = df.loc[
            df.residue_name.isin(STANDARD_AMINO_ACID_MAPPING_1_TO_3.values())
        ]
    df = pd.concat([df] + hets)
    df = sort_dataframe(df)

    df["residue_id"] = (
        df["chain_id"]
        + ":"
        + df["residue_name"]
        + ":"
        + df["residue_number"].astype(str)
    )
    if keep_insertions:
        df["residue_id"] = df.residue_id + ":" + df.insertion

    out = Data(
        coords=protein_df_to_tensor(
            df,
            atoms_to_keep=atom_types,
            fill_value=fill_value_coords,
        ),
        residues=get_sequence(
            df,
            chains=chain_selection,
            insertions=keep_insertions,
            list_of_three=True,
        ),
        id=id,
        residue_id=get_residue_id(df),
        residue_type=residue_type_tensor(df),
        chains=protein_df_to_chain_tensor(df),
    )

    if store_het:
        out.hetatms = [het_data]

    if store_bfactor:
        # group by residue_id and average b_factor per residue
        residue_bfactors = df.groupby("residue_id")["b_factor"].mean(
            numeric_only=True
        )
        out.bfactor = torch.from_numpy(residue_bfactors.values)

    return out


def protein_df_to_chain_tensor(
    df: pd.DataFrame,
    chains_to_keep: Optional[List[str]] = None,
    insertions: bool = True,
    one_hot: bool = False,
    dtype: torch.dtype = torch.int64,
    device: torch.device = torch.device("cpu"),
    per_atom: bool = False,
) -> torch.Tensor:
    """Returns a tensor of chain IDs for a protein structure.

    :param df: DataFrame of protein structure. Must have a column called
        ``"chain_id"`` (and ``insertion`` if the ``insertions=True``).
    :type df: pd.DataFrame
    :param chains_to_keep: List of chains to retain, defaults to ``None``
        (all chains).
    :type chains_to_keep: Optional[List[str]], optional
    :param insertions: Whether or not to keep insertions, defaults to ``True``
    :type insertions: bool, optional
    :param one_hot: Whether or not to return a one-hot encoded tensor
        (``L x num_chains``). If ``False`` an integer tensor is returned.
        Defaults to ``False``.
    :type one_hot: bool, optional
    :return: One hot encoded or integer tensor indicating chain membership for
        each residue.
    :rtype: torch.Tensor
    """
    if "residue_id" not in df.columns:
        get_residue_id(df, insertions=insertions)

    # Select chains to keep from user input
    if chains_to_keep is not None:
        df = df.loc[df.chain_id.isin(chains_to_keep)]

    # Keep or remove insertions
    if not insertions:
        df = df.loc[df.insertion.isin(["", " "])]

    if not per_atom:
        chains = pd.Series(df.residue_id.unique())
        chains = chains.str.split(":").str[0]
    else:
        chains = df.chain_id

    # One-hot encode chain IDs
    chains = pd.get_dummies(chains)
    chains = torch.tensor(chains.values, dtype=dtype, device=device)

    # Integers instead of one-hot
    if not one_hot:
        chains = torch.argmax(chains, dim=1)

    return chains


def protein_df_to_tensor(
    df: pd.DataFrame,
    atoms_to_keep: List[str] = PROTEIN_ATOMS,
    insertions: bool = True,
    fill_value: float = 1e-5,
) -> AtomTensor:
    """
    Transforms a DataFrame of a protein structure into a
    ``Length x Num_Atoms (default 37) x 3`` tensor.

    :param df: DataFrame of protein structure.
    :type df: pd.DataFrame
    :param atoms_to_keep: List of atom types to retain in the tensor.
        Defaults to :const:`~graphein.protein.resi_atoms.PROTEIN_ATOMS`
    :type atoms_to_keep: List[str]
    :param insertions: Whether or not to keep insertions. Defaults to ``True``.
    :type insertions: bool
    :param fill_value: Value to fill missing entries with. Defaults to ``1e-5``.
    :type fill_value: float
    :returns: ``Length x Num_Atoms (default 37) x 3`` tensor.
    :rtype: graphein.protein.tensor.types.AtomTensor
    """
    num_residues = get_protein_length(df, insertions=insertions)
    df = df.loc[df["atom_name"].isin(atoms_to_keep)]
    residue_indices = pd.factorize(
        pd.Series(get_residue_id(df, unique=False))
    )[0]
    atom_indices = df["atom_name"].map(lambda x: atoms_to_keep.index(x)).values

    positions: AtomTensor = (
        torch.zeros((num_residues, len(atoms_to_keep), 3)) + fill_value
    )
    positions[residue_indices, atom_indices] = torch.tensor(
        df[["x_coord", "y_coord", "z_coord"]].values
    ).float()

    return positions


def to_dataframe(
    x: AtomTensor,
    fill_value: float = 1e-5,
    residue_types: Optional[List[str]] = None,
    chains: Optional[Union[List[Union[str, int]], torch.Tensor]] = None,
    insertions: Optional[List[Union[str, float]]] = None,
    b_factors: Optional[Union[List[float], torch.Tensor]] = None,
    occupancy: Optional[Union[List[float], torch.Tensor]] = None,
    charge: Optional[List[int]] = None,
    alt_loc: Optional[List[Union[str, int]]] = None,
    segment_id: Optional[List[Union[str, int]]] = None,
    biopandas: bool = False,
) -> Union[pd.DataFrame, PandasPdb]:
    """Converts an ``AtomTensor`` to a DataFrame.

    ``AtomTensors`` are not a full specification of a structure so missing
    values can be manually provided as arguments - otherwise default values are
    used.

    .. code-block:: python

        import graphein.protein.tensor as gpt

        protein = gpt.Protein().from_pdb_code("3eiy")
        to_dataframe(protein.coords)

    .. seealso::

        :class:`graphein.protein.tensor.types.AtomTensor`
        :meth:`graphein.protein.tensor.io.to_pdb`
        :meth:`graphein.protein.tensor.sequence.infer_residue_types`


    :param x: AtomTensor to convert (Shape: ``num residues x 37 x 3``).
    :type x: AtomTensor
    :param fill_value: Fill value used to denote missing atoms in the
    ``AtomTensor``, defaults to ``1e-5``.
    :type fill_value: float, optional
    :param residue_types: List of three-letter residue IDs (length: num residues
        ), defaults to ``None`` (inferred from ``AtomTensor``; this may break
        for incomplete structures with missing atoms.)
    :type residue_types: Optional[List[str]], optional
    :param chains: List or tensor of chain IDs, defaults to ``None`` (``"A"``
        for all residues).
    :type chains: Optional[Union[List[Union[str, int]], torch.Tensor]], optional
    :param insertions: List of insertion codes, defaults to ``None`` (``""``).
    :type insertions: Optional[List[Union[str, float]]], optional
    :param b_factors: List or tensor of b factors (length: num residues),
        defaults to ``None`` (``""``). If ``b_factors`` is of length/shape
        number of residues (as opposed to number of atoms) it is automatically
        unravelled to the correct length.
    :type b_factors: Optional[List[Union[str, float]]], optional
    :param occupancy: List or tensor of occupancy values (length: num residues),
        defaults to ``None`` (``1.0``).
    :type occupancy: Optional[List[Union[str, float]]], optional
    :param charge: List or or tensor of atom charges, defaults to ``None``
        (``"0"``).
    :type charge: Optional[List[int]], optional
    :param alt_loc: List or tensor of alt_loc codes, defaults to ``None``
        (``""``)
    :type alt_loc: Optional[List[Union[str, int]]], optional
    :param segment_id: List or tensor of segment IDs, defaults to ``None``
        (``""``).
    :type segment_id: Optional[List[Union[str, int]]], optional
    :param biopandas: Whether to return a ``pd.DataFrame`` or ``BioPandas``
        ``PandasPdb`` object, defaults to ``False`` (``pd.DataFrame``).
    :type biopandas: bool, optional
    :return: ``DataFrame`` or ``PandasPdb object``.
    :rtype: Union[pd.DataFrame, PandasPdb]
    """
    nz = (x - fill_value).nonzero()
    nz = nz[torch.where(nz[:, 2] == 0)]
    res_nums = nz[:, 0] + 1

    atom_number = np.arange(1, len(res_nums) + 1)
    numbering_map = {v: k for k, v in ATOM_NUMBERING.items()}
    atom_type = nz[:, 1]
    atom_type = [numbering_map[a.item()] for a in atom_type]

    if residue_types is None:
        residue_types = infer_residue_types(
            x, fill_value=fill_value, return_sequence=False
        )
    if isinstance(residue_types, torch.Tensor):
        residue_types = [
            STANDARD_AMINO_ACID_MAPPING_1_TO_3[STANDARD_AMINO_ACIDS[a]]
            for a in residue_types
        ]
    residue_types = [residue_types[a - 1] for a in res_nums]
    element_symbols = [ELEMENT_SYMBOL_MAP[a] for a in atom_type]

    chains = ["A"] * len(res_nums) if chains is None else chains[res_nums - 1]
    if b_factors is not None:
        num_b_factors = (
            len(b_factors)
            if isinstance(b_factors, list)
            else b_factors.shape[0]
        )
        b_factors = (
            b_factors[res_nums - 1]
            if num_b_factors == x.shape[0]
            else b_factors
        )
        if isinstance(b_factors, torch.Tensor):
            b_factors = b_factors.tolist()
    else:
        b_factors = [0.0] * len(res_nums)
    if segment_id is None:
        segment_id = [""] * len(res_nums)
    if insertions is None:
        insertions = [""] * len(res_nums)
    if occupancy is None:
        occupancy = [1.0] * len(res_nums)
    if charge is None:
        charge = [0] * len(res_nums)
    if alt_loc is None:
        alt_loc = [""] * len(res_nums)

    blank_1 = [""] * len(res_nums)
    blank_2 = [""] * len(res_nums)
    blank_3 = [""] * len(res_nums)
    blank_4 = [""] * len(res_nums)

    # NB brittle, bad assumption; may break
    record_names = ["ATOM" if i < 37 else "HETATM" for i in nz[:, 1]]
    coords = get_full_atom_coords(x)[0]

    out = {
        "record_name": record_names,
        "atom_number": atom_number,
        "blank_1": blank_1,
        "atom_name": atom_type,
        "alt_loc": alt_loc,
        "residue_name": residue_types,
        "blank_2": blank_2,
        "chain_id": chains,
        "residue_number": res_nums,
        "insertion": insertions,
        "blank_3": blank_3,
        "x_coord": coords[:, 0],
        "y_coord": coords[:, 1],
        "z_coord": coords[:, 2],
        "occupancy": occupancy,
        "b_factor": b_factors,
        "blank_4": blank_4,
        "segment_id": segment_id,
        "element_symbol": element_symbols,
        "charge": charge,
        "line_idx": atom_number,
    }
    df = pd.DataFrame().from_dict(out)
    if biopandas:
        ppdb = PandasPdb()
        ppdb.df["ATOM"] = df
        return ppdb
    return df


def to_pdb(x: AtomTensor, out_path: str, gz: bool = False, **kwargs):
    """
    Writes an ``AtomTensor`` to a PDB file.

    .. seealso::

        :class:`graphein.protein.tensor.types.AtomTensor`
        :func:`graphein.protein.tensor.to_dataframe`

    :param x: ``AtomTensor`` describing protein structure to write.
    :type x: AtomTensor
    :param out_path: Path to output pdb file.
    :type out_path: str
    :param gz: Whether to gzip out the output, defaults to ``False``.
    :type gz: bool, optional
    :param kwargs: Keyword args for :func:`graphein.protein.tensor.to_dataframe`
    """
    df = to_dataframe(x, **kwargs)
    ppdb = PandasPdb()
    ppdb.df["ATOM"] = df
    ppdb.to_pdb(path=out_path, gz=gz, append_newline=True)



from typing import Union

import numpy as np
import torch
from torch_geometric.utils import unbatch
from torchmetrics import Metric

from ...protein.tensor.types import AtomTensor, CoordTensor


def tm_score(
    y_hat: Union[CoordTensor, AtomTensor], y: Union[CoordTensor, AtomTensor]
) -> torch.Tensor:
    """Compute TMScore between ``y_hat`` and ``y``. Requires aligned structures.

    TM-score is a measure of similarity between two protein structures.
    The TM-score is intended as a more accurate measure of the global
    similarity of full-length protein structures than the often used RMSD
     measure. The TM-score indicates the similarity between two structures
    by a score between ``[0, 1]``, where 1 indicates a perfect match
    between two structures (thus the higher the better). Generally scores
    below 0.20 corresponds to randomly chosen unrelated proteins whereas
    structures with a score higher than 0.5 assume roughly the same fold.
    A quantitative study shows that proteins of TM-score = 0.5 have a
    posterior probability of 37% in the same CATH topology family and of
    13% in the same SCOP fold family. The probabilities increase rapidly
    when TM-score > 0.5. The TM-score is designed to be independent of
    protein lengths.

    https://en.wikipedia.org/wiki/Template_modeling_score

    :param y_hat: Tensor of atom positions (aligned to ``y``).
    :type y_hat: Union[graphein.protein.tensor.types.CoordTensor,
        graphein.protein.tensor.types.AtomTensor]
    :param y: Tensor of groundtruth/reference atom positions.
    :type y: Union[graphein.protein.tensor.types.CoordTensor,
        graphein.protein.tensor.types.AtomTensor]
    :returns: TMScore of aligned pair.
    :rtype: torch.Tensor
    """
    # Get CA
    if y_hat.ndim == 3:
        y_hat = y_hat[:, 1, :]
    if y.ndim == 3:
        y = y[:, 1, :]

    l_target = y.shape[0]

    d0_l_target = 1.24 * np.power(l_target - 15, 1 / 3) - 1.8

    di = torch.pairwise_distance(y_hat, y)

    return torch.sum(1 / (1 + (di / d0_l_target) ** 2)) / l_target


class TMScore(Metric):
    def __init__(self):
        super().__init__()
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state(
            "correct", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )
        higher_is_better: bool = True
        full_state_update: bool = True

    @property
    def higher_is_better(self):
        return True

    def update(
        self,
        pred: Union[CoordTensor, AtomTensor],
        target: Union[CoordTensor, AtomTensor],
        batch: torch.Tensor,
    ):
        y = unbatch(target, batch)
        y_hat = unbatch(pred, batch)

        for i, j in zip(y, y_hat):
            self.correct += tm_score(i, j)
            self.total += 1

    def compute(self):
        return self.correct / self.total



"""Tests for PyTorch Geometric Dataset constructors."""

import os
import shutil
from pathlib import Path

import pytest
from numpy.testing import assert_array_equal

import graphein.protein as gp
from graphein.ml import GraphFormatConvertor

try:
    import torch_geometric

    PYG_AVAIL = True
except ImportError:
    PYG_AVAIL = False

try:
    from graphein.ml import (
        InMemoryProteinGraphDataset,
        ProteinGraphDataset,
        ProteinGraphListDataset,
    )
except (NameError, ImportError):
    pass

ROOT_DIR = Path(__file__).parent


@pytest.mark.skipif(not PYG_AVAIL, reason="PyG not installed")
def test_list_dataset():
    # Construct graphs
    graphs = gp.construct_graphs_mp(
        pdb_code_it=["3eiy", "4hhb", "1lds", "2ll6"], return_dict=False
    )

    # do some transformation
    graphs = [gp.extract_subgraph_from_chains(g, ["A"]) for g in graphs]

    # Convert to PyG Data format
    convertor = GraphFormatConvertor(src_format="nx", dst_format="pyg")
    graphs = [convertor(g) for g in graphs]

    # Create dataset
    ds = ProteinGraphListDataset(
        root=ROOT_DIR, data_list=graphs, name="list_test"
    )

    assert len(ds) == len(graphs)
    assert os.path.exists(ROOT_DIR / "processed" / "data_list_test.pt")

    for i, d in enumerate(ds):
        assert_array_equal(d.edge_index, graphs[i].edge_index)
        assert d.node_id == graphs[i].node_id
        assert_array_equal(d.coords[0], graphs[i].coords[0])
        assert d.name == graphs[i].name
        assert d.num_nodes == graphs[i].num_nodes
    # Clean up
    shutil.rmtree(ROOT_DIR / "processed")


@pytest.mark.skipif(not PYG_AVAIL, reason="PyG not installed")
def test_in_memory_dataset():
    pdb_list = ["7vxg", "7dys", "7whv", "7eek"]
    uniprots = ["P10513", "B1VC86", "P13948", "P17998"]

    ds = InMemoryProteinGraphDataset(
        root=ROOT_DIR,
        name="in_memory_test",
        pdb_codes=pdb_list,
        uniprot_ids=uniprots,
    )
    assert len(ds) == len(pdb_list + uniprots)

    # Check raw files exist
    for pdb in pdb_list + uniprots:
        assert f"{pdb}.pdb" in os.listdir(ROOT_DIR / "raw")

    # Check processed data exists
    assert os.path.exists(ROOT_DIR / "processed" / "data_in_memory_test.pt")

    # Clean up
    shutil.rmtree(ROOT_DIR / "raw")
    shutil.rmtree(ROOT_DIR / "processed")


@pytest.mark.skipif(not PYG_AVAIL, reason="PyG not installed")
def test_protein_graph_dataset():
    pdb_list = ["4rty", "4r01", "5e08", "6f15"]
    uniprots = ["A0A0A1EI90", "A0A0B4JCS5", "A0A0B4JCZ3", "A0A0B4JCZ0"]

    ds = ProteinGraphDataset(
        root=ROOT_DIR,
        pdb_codes=pdb_list,
        uniprot_ids=uniprots,
    )
    assert len(ds) == len(pdb_list + uniprots)

    # Check raw files and processed data exist
    for pdb in pdb_list + uniprots:
        assert f"{pdb}.pdb" in os.listdir(ROOT_DIR / "raw")
        assert f"{pdb}.pt" in os.listdir(ROOT_DIR / "processed")

    # Clean up
    shutil.rmtree(ROOT_DIR / "raw")
    shutil.rmtree(ROOT_DIR / "processed")



"""Nussinov algorithm for computing RNA secondary structure adopted from
adopted from https://github.com/cgoliver/Nussinov/blob/master/nussinov.py"""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ryan Greenhalgh
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Tuple

import numpy as np

MIN_LOOP_LENGTH = 10


def pair_check(
    base_pair_tuple: Tuple[str, str],
) -> bool:
    """Check if there is a base pair interaction between two bases.

    :param base_pair_tuple: Base pair tuple to test.
    :type base_pair_tuple: Tuple[str, str]
    :return: ``True`` if base pairs interact else ``False``.
    :rtype: bool
    """
    return base_pair_tuple in [("A", "U"), ("U", "A"), ("C", "G"), ("G", "C")]


def optimal_pairing(
    i: int,
    j: int,
    sequence: str,
) -> int:
    """Find the optimal pairing between two positions.

    :param i: Position ``i`` in sequence.
    :type i: int
    :param j: Position ``j`` in sequence.
    :type j: int
    :param sequence: RNA sequence.
    :type sequence: str
    :return:
    :rtype: int
    """

    # base case: no pairs allowed when i and j are less than 4 bases apart
    if i >= j - MIN_LOOP_LENGTH:
        return 0

    # i and j can either be paired or not be paired, if not paired then the optimal score is optimal_pairing(i,j-1)
    unpaired = optimal_pairing(i, j - 1, sequence)

    # check if j can be involved in a pairing with a position t
    pairing = [
        1
        + optimal_pairing(i, t - 1, sequence)
        + optimal_pairing(t + 1, j - 1, sequence)
        for t in range(i, j - MIN_LOOP_LENGTH)
        if pair_check((sequence[t], sequence[j]))
    ] or [0]
    paired = max(pairing)

    return max(unpaired, paired)


def traceback(
    i: int,
    j: int,
    structure: str,
    DP: np.ndarray,
    sequence: str,
) -> None:
    """Recursively check pairing and interactions between base pairs.

    :param i: Position ``i`` in sequence.
    :type i: int
    :param j: Position ``j`` in sequence.
    :type j: int
    :param structure: Dot-bracket notation for RNA seq.
    :type structure: str
    :param DP: Numpy matrix to cache
    :type DP: np.ndarray
    :param sequence: RNA sequence
    :type sequence: str
    """

    # in this case we've gone through the whole sequence. Nothing to do.
    if j <= i:
        return
    # if j is unpaired, there will be no change in score when we take it out, so we just recurse to the next index
    elif DP[i][j] == DP[i][j - 1]:
        traceback(i, j - 1, structure, DP, sequence)
    # consider cases where j forms a pair.
    else:
        # try pairing j with a matching index k to its left.
        for k in [
            b
            for b in range(i, j - MIN_LOOP_LENGTH)
            if pair_check((sequence[b], sequence[j]))
        ]:
            # if the score at i,j is the result of adding 1 from pairing (j,k) and whatever score
            # comes from the substructure to its left (i, k-1) and to its right (k+1, j-1)
            if k < 1:
                if DP[i][j] == DP[k + 1][j - 1] + 1:
                    structure.append((k, j))
                    traceback(k + 1, j - 1, structure, DP, sequence)
                    break
            elif DP[i][j] == DP[i][k - 1] + DP[k + 1][j - 1] + 1:
                # add the pair (j,k) to our list of pairs
                structure.append((k, j))
                # move the recursion to the two substructures formed by this pairing
                traceback(i, k - 1, structure, DP, sequence)
                traceback(k + 1, j - 1, structure, DP, sequence)
                break


def write_structure(
    sequence: str,
    structure: str,
) -> str:
    """Convert structure to string.

    :param sequence: RNA sequence.
    :type sequence: str
    :param structure: RNA dot-bracket.
    :type structure: str
    :return: Dot-bracket notation as a string.
    :rtype: str
    """
    dot_bracket = ["." for _ in range(len(sequence))]
    for s in structure:
        dot_bracket[min(s)] = "("
        dot_bracket[max(s)] = ")"
    return "".join(dot_bracket)


def initialize(
    N: int,
) -> np.ndarray:
    """Initialize DP matrix. ``NxN`` matrix that stores the scores of the optimal pairings.

    :param N: Length of RNA sequence
    :type N: int
    :return: DP matrix for Nussinov Algorithm
    :rtype: np.ndarray
    """

    DP = np.empty((N, N))
    DP[:] = np.nan
    for k in range(MIN_LOOP_LENGTH):
        for i in range(N - k):
            j = i + k
            DP[i][j] = 0
    return DP


def nussinov(
    sequence: str,
) -> str:
    """Nussinov algorithm for predicting RNA ss in dot-bracket notation.

    :param sequence: RNA sequence.
    :type sequence: str
    :return: Dot-bracket notation for RNA sequence.
    :rtype: str
    """
    N = len(sequence)
    DP = initialize(N)
    structure = []

    # fill the DP matrix diagonally
    for k in range(MIN_LOOP_LENGTH, N):
        for i in range(N - k):
            j = i + k
            DP[i][j] = optimal_pairing(i, j, sequence)

    # copy values to lower triangle to avoid null references
    for i in range(N):
        for j in range(i):
            DP[i][j] = DP[j][i]

    traceback(0, N - 1, structure, DP, sequence)

    return write_structure(sequence, structure)



"""Unit test for the graphein package."""

import sys

import pytest

# Import package, test suite, and other packages as needed
import graphein


def test_sidechainnet_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "graphein" in sys.modules



"""Featurization functions for graph nodes using DSSP predicted features."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
import re
import tempfile
from typing import Any, Dict, Optional

import networkx as nx
import pandas as pd
from Bio.PDB.DSSP import dssp_dict_from_pdb_file, residue_max_acc
from loguru import logger

from graphein.protein.resi_atoms import STANDARD_AMINO_ACID_MAPPING_1_TO_3
from graphein.protein.utils import save_pdb_df_to_pdb
from graphein.utils.dependencies import is_tool, requires_external_dependencies

DSSP_COLS = [
    "chain",
    "resnum",
    "icode",
    "aa",
    "ss",
    "asa",
    "phi",
    "psi",
    "dssp_index",
    "NH_O_1_relidx",
    "NH_O_1_energy",
    "O_NH_1_relidx",
    "O_NH_1_energy",
    "NH_O_2_relidx",
    "NH_O_2_energy",
    "O_NH_2_relidx",
    "O_NH_2_energy",
]
"""Columns in DSSP Output."""

DSSP_SS = ["H", "B", "E", "G", "I", "T", "S"]
"""Secondary structure types detected by ``DSSP``"""


def parse_dssp_df(dssp: Dict[str, Any]) -> pd.DataFrame:
    """
    Parse ``DSSP`` output to DataFrame.

    :param dssp: Dictionary containing ``DSSP`` output
    :type dssp: Dict[str, Any]
    :return: pd.DataFrame containing parsed ``DSSP`` output
    :rtype: pd.DataFrame
    """
    appender = []
    for k in dssp[1]:
        to_append = []
        y = dssp[0][k]
        chain = k[0]
        residue = k[1]
        # het = residue[0]
        resnum = residue[1]
        icode = residue[2]
        to_append.extend([chain, resnum, icode])
        to_append.extend(y)
        appender.append(to_append)

    return pd.DataFrame.from_records(appender, columns=DSSP_COLS)


@requires_external_dependencies("mkdssp")
def add_dssp_df(
    G: nx.Graph,
    dssp_config: Optional[DSSPConfig],
) -> nx.Graph:
    """
    Construct DSSP dataframe and add as graph level variable to protein graph.

    :param G: Input protein graph
    :param G: nx.Graph
    :param dssp_config: DSSPConfig object. Specifies which executable to run.
        Located in `:obj:graphein.protein.config`.
    :type dssp_config: DSSPConfig, optional
    :return: Protein graph with DSSP dataframe added
    :rtype: nx.Graph
    """
    # if dssp_config is None:
    config = G.graph["config"]
    pdb_code = G.graph["pdb_code"]
    path = G.graph["path"]
    pdb_name = G.graph["name"]

    # Extract DSSP executable
    executable = dssp_config.executable

    # Ensure that DSSP is on PATH and is marked as an executable.
    assert is_tool(
        executable
    ), "DSSP must be on PATH and marked as an executable"

    pdb_file = None
    if path:
        if os.path.isfile(path):
            pdb_file = path
    else:
        if config.pdb_dir:
            if os.path.isfile(config.pdb_dir / (pdb_code + ".pdb")):
                pdb_file = config.pdb_dir / (pdb_code + ".pdb")

    # get dssp version string
    dssp_version = re.search(
        r"version ([\d\.]+)", os.popen(f"{executable} --version").read()
    ).group(
        1
    )  # e.g. "4.0.4"
    # Check for existence of pdb file. If not, reconstructs it from the raw df.
    if pdb_file:
        dssp_dict = dssp_dict_from_pdb_file(
            pdb_file, DSSP=executable, dssp_version=dssp_version
        )
    else:
        with tempfile.TemporaryDirectory() as tmpdirname:
            save_pdb_df_to_pdb(
                G.graph["raw_pdb_df"], tmpdirname + f"/{pdb_name}.pdb"
            )
            dssp_dict = dssp_dict_from_pdb_file(
                tmpdirname + f"/{pdb_name}.pdb",
                DSSP=executable,
                dssp_version=dssp_version,
            )

    if len(dssp_dict[0]) == 0:
        raise ValueError(
            "DSSP could not be calculated. Check DSSP version "
            f"({dssp_version}) orthat the input PDB file is valid."
        )

    if config.verbose:
        logger.debug(f"Using DSSP executable '{executable}'")

    dssp_dict = parse_dssp_df(dssp_dict)
    # Convert 1 letter aa code to 3 letter
    dssp_dict["aa"] = dssp_dict["aa"].map(STANDARD_AMINO_ACID_MAPPING_1_TO_3)

    # Resolve UNKs
    # NOTE: the original didn't work if HETATM residues exist in DSSP output
    _raw_pdb_df = G.graph["raw_pdb_df"].copy().drop_duplicates("node_id")
    _dssp_df_unk = dssp_dict.loc[dssp_dict["aa"] == "UNK"][
        ["chain", "resnum", "icode"]
    ]
    for chain, resnum, icode in _dssp_df_unk.values:
        dssp_dict.loc[
            (dssp_dict["chain"] == chain)
            & (dssp_dict["resnum"] == resnum)  # e.g. 'H'
            & (dssp_dict["icode"] == icode),  # e.g. 100  # e.g. 'E' or ' '
            "aa",
        ] = _raw_pdb_df.loc[
            (_raw_pdb_df["chain_id"] == chain)
            & (_raw_pdb_df["residue_number"] == resnum)
            & (
                _raw_pdb_df["insertion"] == icode.strip()
            )  # why strip?: dssp icode is a space, raw_pdb_df is empty string
        ][
            "residue_name"
        ].values[
            0
        ]

    # Construct node IDs
    dssp_dict["node_id"] = (
        dssp_dict["chain"]
        + ":"
        + dssp_dict["aa"]
        + ":"
        + dssp_dict["resnum"].astype(str)
    )
    if G.graph["config"].insertions:
        dssp_dict["node_id"] = (
            dssp_dict["node_id"] + ":" + dssp_dict["icode"].apply(str)
        )
        # Replace trailing : for non insertions
        dssp_dict["node_id"] = dssp_dict["node_id"].str.replace(
            r":\s*$", "", regex=True
        )

    dssp_dict.set_index("node_id", inplace=True)

    if config.verbose:
        logger.debug(dssp_dict)

    # Assign DSSP Dict
    G.graph["dssp_df"] = dssp_dict

    return G


def add_dssp_feature(G: nx.Graph, feature: str) -> nx.Graph:
    """
    Adds specified amino acid feature as calculated
    by DSSP to every node in a protein graph

    :param G: Protein structure graph to add dssp feature to
    :type G: nx.Graph
    :param feature: string specifying name of DSSP feature to add:
        ``"chain"``,
        ``"resnum"``,
        ``"icode"``,
        ``"aa"``,
        ``"ss"``,
        ``"asa"``,
        ``"phi"``,
        ``"psi"``,
        ``"dssp_index"``,
        ``"NH_O_1_relidx"``,
        ``"NH_O_1_energy"``,
        ``"O_NH_1_relidx"``,
        ``"O_NH_1_energy"``,
        ``"NH_O_2_relidx"``,
        ``"NH_O_2_energy"``,
        ``"O_NH_2_relidx"``,
        ``"O_NH_2_energy"``,
        These names are accessible in the DSSP_COLS list
    :type feature: str
    :return: Protein structure graph with DSSP feature added to nodes
    :rtype: nx.Graph
    """
    if "dssp_df" not in G.graph:
        G = add_dssp_df(G, G.graph["config"].dssp_config)

    config = G.graph["config"]
    dssp_df = G.graph["dssp_df"]

    # Change to not allow for atom granularity?
    if config.granularity == "atom":
        # TODO confirm below is not needed and remove
        """
        # If granularity is atom, apply residue feature to every atom
        for n in G.nodes():
            residue = n.split(":")
            residue = residue[0] + ":" + residue[1] + ":" + residue[2]

            G.nodes[n][feature] = dssp_df.loc[residue, feature]
        """
        raise NameError(f"DSSP residue features ({feature}) \
            cannot be added to atom granularity graph")

    else:
        nx.set_node_attributes(G, dict(dssp_df[feature]), feature)

    if config.verbose:
        logger.debug("Added " + feature + " features to graph nodes")

    return G


def rsa(G: nx.Graph) -> nx.Graph:
    """
    Adds RSA (relative solvent accessibility) of each residue in protein graph
    as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with rsa values added
    :rtype: nx.Graph
    """

    # Calculate RSA
    try:
        dssp_df = G.graph["dssp_df"]
    except KeyError:
        G = add_dssp_df(G, G.graph["config"].dssp_config)
        dssp_df = G.graph["dssp_df"]
    dssp_df["max_acc"] = dssp_df["aa"].map(residue_max_acc["Sander"].get)
    dssp_df[["asa", "max_acc"]] = dssp_df[["asa", "max_acc"]].astype(float)
    dssp_df["rsa"] = dssp_df["asa"] / dssp_df["max_acc"]

    G.graph["dssp_df"] = dssp_df

    return add_dssp_feature(G, "rsa")


def asa(G: nx.Graph) -> nx.Graph:
    """
    Adds ASA of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with asa values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "asa")


def phi(G: nx.Graph) -> nx.Graph:
    """
    Adds phi-angles of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with phi-angles values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "phi")


def psi(G: nx.Graph) -> nx.Graph:
    """
    Adds psi-angles of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with psi-angles values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "psi")


def secondary_structure(G: nx.Graph) -> nx.Graph:
    """
    Adds secondary structure of each residue in protein graph
    as calculated by DSSP in the form of a string

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with secondary structure added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "ss")



"""Base Config object for use with Protein Graph Construction."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
from functools import partial
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

from deepdiff import DeepDiff
from pydantic import BaseModel, validator

from graphein.protein.edges.distance import add_peptide_bonds
from graphein.protein.features.nodes.amino_acid import meiler_embedding
from graphein.utils.config import PartialMatchOperator, PathMatchOperator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class DSSPConfig(BaseModel):
    executable: str = "mkdssp"


class GetContactsConfig(BaseModel):
    """
    Config object for parameters relating to running ``GetContacts``.
    ``GetContacts`` is an optional dependency from which intramolecular
    interactions can be computed and used as edges in the graph.

    More information about ``GetContacts`` can be found at
    https://getcontacts.github.io/

    :param get_contacts_path: Path to ``GetContacts`` installation
    :type get_contacts_path: pathlib.Path
    :param contacts_dir: Path to store output of ``GetContacts``
    :type contacts_dir: pathlib.Path
    :param pdb_dir: Path to PDB files to be used to compute intramolecular
        interactions.
    :type pdb_dir: pathlib.Path
    :param granularity: Specifies the node types of the graph, defaults to
        ``"CA"`` for alpha-carbons as nodes. Other options are ``"CB"``
        (beta-carbon), ``"atom"`` for all-atom graphs, and ``"centroid"``
        for nodes positioned as residue centroids.
    :type granularity: str
    """

    get_contacts_path: Path = Path(
        "/Users/arianjamasb/github/getcontacts/"
    ).resolve()  # TODO: get rid of this absolute path
    contacts_dir: Path = Path(
        os.path.join(os.path.dirname(__file__), "../examples/contacts/")
    ).resolve()
    pdb_dir: Path = Path(
        os.path.join(os.path.dirname(__file__), "../examples/pdbs/")
    ).resolve()
    granularity: str = "CA"


GraphAtoms = Literal[
    "N",
    "CA",
    "C",
    "O",
    "CB",
    "OG",
    "CG",
    "CD1",
    "CD2",
    "CE1",
    "CE2",
    "CZ",
    "OD1",
    "ND2",
    "CG1",
    "CG2",
    "CD",
    "CE",
    "NZ",
    "OD2",
    "OE1",
    "NE2",
    "OE2",
    "OH",
    "NE",
    "NH1",
    "NH2",
    "OG1",
    "SD",
    "ND1",
    "SG",
    "NE1",
    "CE3",
    "CZ2",
    "CZ3",
    "CH2",
    "OXT",
]
"""Allowable atom types for nodes in the graph."""

GranularityOpts = Literal["atom", "centroids"]
"""Allowable granularity options for nodes in the graph."""

AltLocsOpts = Union[
    bool,
    Literal[
        "max_occupancy", "min_occupancy", "first", "last", "exclude", "include"
    ],
]
"""Allowable altlocs options for alternative locations handling."""


class ProteinGraphConfig(BaseModel):
    """
    Config Object for Protein Structure Graph Construction.

    If you encounter a problematic structure, perusing
    https://www.umass.edu/microbio/chime/pe_beta/pe/protexpl/badpdbs.htm
    may provide some additional insight. PDBs are notoriously troublesome and
    this is an excellent overview.

    :param granularity: Controls the granularity of the graph construction.
        ``"atom"`` builds an atomic-scale graph where nodes are constituent
        atoms. Residue-level graphs can be build by specifying which constituent
        atom should represent node positions
        (see :const:`~graphein.protein.config.GraphAtoms`). Additionally,
        ``"centroids"`` can be specified to
        compute the centre of gravity for a given atom (Specified in
        :const:`~graphein.protein.config.GranularityOpts`). Defaults to
        ``"CA"`` (alpha-Carbon).
    :type granularity: str (Union[graphein.protein.config.GraphAtoms,
        graphein.protein.config.GranularityOpts])
    :param keep_hets: Controls whether or not heteroatoms are removed from the
        PDB file. These are typically modified residues, bound ligands,
        crystallographic adjuvants, ions or water molecules. For more
        information, see: https://proteopedia.org/wiki/index.php/Hetero_atoms
    :type keep_hets: List[str]
    :param insertions: Controls whether or not insertions are allowed.
    :type insertions: bool
    :param alt_locs: Controls whether or not alternative locations are allowed. The supported values are
        ``"max_occupancy"``, ``"min_occupancy"``, ``"first"``, ``"last"``, ``"exclude"``. First two will leave altlocs
        with the highest/lowest occupancies, next two will leave first/last in the PDB file ordering. The ``"exclude"``
        value will drop them entirely and ``"include"`` leave all of them. Additionally, boolean values are the aliases
        for the latest options. Default is ``"max_occupancy"``.
    :type alt_locs: AltLocsOpts
    :param pdb_dir: Specifies path to download protein structures into.
    :type pdb_dir: pathlib.Path. Optional.
    :param verbose: Specifies verbosity of graph creation process.
    :type verbose: bool
    :param exclude_waters: Specifies whether or not water molecules are excluded
        from the structure
    :type excluded_waters: bool
    :param deprotonate: Specifies whether or not to remove ``H`` atoms from the
        graph.
    :type deprotonate: bool
    :param protein_df_processing_functions: List of functions that take a
        ``pd.DataFrame`` and return a ``pd.DataFrame``. This allows users to
        define their own series of processing functions for the protein
        structure DataFrame and override the default sequencing of processing
        steps provided by Graphein. We refer users to our low-level API
        tutorial for more details.
    :type protein_df_processing_functions: Optional[List[Callable]]
    :param edge_construction_functions: List of functions that take an
        ``nx.Graph`` and return an ``nx.Graph`` with desired edges added.
        Prepared edge constructions can be found in
        :ref:`graphein.protein.edges`.
    :type edge_construction_functions: List[Callable]
    :param node_metadata_functions: List of functions that take an ``nx.Graph``
    :type node_metadata_functions: List[Callable], optional
    :param edge_metadata_functions: List of functions that take an
    :type edge_metadata_functions: List[Callable], optional
    :param graph_metadata_functions: List of functions that take an ``nx.Graph``
        and return an ``nx.Graph`` with added graph-level features and metadata.
    :type graph_metadata_functions: List[Callable], optional
    :param get_contacts_config: Config object containing parameters for running
        ``GetContacts`` for computing intramolecular contact-based edges.
        Defaults to ``None``.
    :type get_contacts_config: GetContactsConfig, optional
    :param dssp_config: Config Object containing reference to ``DSSP``
        executable. Defaults to ``None``. **NB** DSSP must be installed. See
        installation instructions:
        https://graphein.ai/getting_started/installation.html#optional-dependencies
    :type dssp_config: DSSPConfig, optional
    """

    granularity: Union[GraphAtoms, GranularityOpts] = "CA"
    keep_hets: List[str] = []
    insertions: bool = True
    alt_locs: AltLocsOpts = "max_occupancy"
    pdb_dir: Optional[Path] = None
    verbose: bool = False
    exclude_waters: bool = True
    deprotonate: bool = False

    # Graph construction functions
    protein_df_processing_functions: Optional[List[Callable]] = None
    edge_construction_functions: List[Union[Callable, str]] = [
        add_peptide_bonds
    ]
    node_metadata_functions: Optional[List[Union[Callable, str]]] = [
        meiler_embedding
    ]
    edge_metadata_functions: Optional[List[Union[Callable, str]]] = None
    graph_metadata_functions: Optional[List[Callable]] = None

    # External Dependency configs
    get_contacts_config: Optional[GetContactsConfig] = None
    dssp_config: Optional[DSSPConfig] = None

    @validator("alt_locs")
    def convert_alt_locs_aliases(cls, v):
        if v is True:
            return "include"
        elif v is False:
            return "exclude"
        else:
            return v

    def __eq__(self, other: Any) -> bool:
        """
        Overwrites the BaseModel __eq__ function in order to check more
        specific cases (like partial functions).
        """
        if isinstance(other, ProteinGraphConfig):
            return (
                DeepDiff(
                    self,
                    other,
                    custom_operators=[
                        PartialMatchOperator(types=[partial]),
                        PathMatchOperator(types=[Path]),
                    ],
                )
                == {}
            )
        return self.dict() == other


class ProteinMeshConfig(BaseModel):
    """
    Config object for parameters relating to Protein Mesh construction with
    ``PyMol``

    **NB** PyMol must be installed. See:
    https://graphein.ai/getting_started/installation.html#optional-dependencies

    :param pymol_command_line_options: List of CLI args for running PyMol.
        See: https://pymolwiki.org/index.php/Command_Line_Options.
        Defaults to ``"-cKq"`` ()
    :type pymol_command_line_options: str, optional
    :param pymol_commands: List of Commands passed to PyMol in surface
        construction.
    :type pymol_commands: List[str], optional
    """

    pymol_command_line_options: Optional[str] = "-cKq"
    pymol_commands: Optional[List[str]] = ["show surface"]



"""Functions for computing atomic structure of molecules."""

import networkx as nx

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from loguru import logger as log


def add_atom_bonds(G: nx.Graph) -> nx.Graph:
    """Adds atomic bonds to a molecular graph.

    :param G: Molecular graph to add atomic bond edges to.
    :type G: nx.Graph
    :return: Molecular graph with atomic bonds added.
    :rtype: nx.Graph
    """
    for bond in G.graph["rdmol"].GetBonds():
        n1, n2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        sym1, sym2 = (
            G.graph["rdmol"].GetAtoms()[n1].GetSymbol(),
            G.graph["rdmol"].GetAtoms()[n2].GetSymbol(),
        )
        n1 = f"{sym1}:{str(n1)}"
        n2 = f"{sym2}:{str(n2)}"
        if G.has_edge(n1, n2):
            G.edges[n1, n2]["kind"].add("bond")
            G.edges[n1, n2]["bond"] = bond
        else:
            G.add_edge(n1, n2, kind={"bond"}, bond=bond)
    return G



"""Utilities for retrieving molecular data from ZINC.

Adapted from smilite (https://github.com/rasbt/smilite) by Sebastian Raschka.
"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT

# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import ssl
from typing import Dict, List
from urllib.error import URLError

from tqdm.rich import tqdm

from graphein.utils.dependencies import import_message

try:
    from smilite import *
except ImportError:
    import_message(
        "graphein.molecule.zinc", "smilite", "smilite", True, extras=True
    )


def get_smiles_from_zinc(zinc_id: str, backend="zinc15") -> str:
    """
    Gets the corresponding SMILE string for a ZINC ID query from
    the ZINC online database (https://zinc.docking.org/). Requires an internet connection.

    :param zinc_id: A valid ZINC ID, e.g. ``'ZINC00029323'``
    :type zinc_id: str
    :param backend: ``zinc12`` or ``zinc15``
    :type backend: str
    :returns: the SMILE string for the corresponding ZINC ID.
        E.g., ``'COc1cccc(c1)NC(=O)c2cccnc2'``
    :rtype: str
    """
    try:
        return smilite.get_zinc_smile(zinc_id, backend=backend)
    except URLError:
        ssl._create_default_https_context = ssl._create_unverified_context
        return smilite.get_zinc_smile(zinc_id, backend=backend)


def get_zinc_id_from_smile(smile: str, backend: str = "zinc15") -> List[str]:
    """
    Gets the corresponding ZINC ID(s) for a SMILE string query from
    the ZINC online database. Requires an internet connection.

    :param smile_str: A valid SMILE string, e.g.,
            ``C[C@H]1CCCC[NH+]1CC#CC(c2ccccc2)(c3ccccc3)O'``
    :type smile_str: str
    :param backend:  Specifies the database backend, ``"zinc12"`` or ``"zinc15"``.
    :type backend: str
    :returns: the SMILE string for the corresponding ZINC ID(s) in a list.
        E.g., ``['ZINC01234567', 'ZINC01234568', 'ZINC01242053', 'ZINC01242055']``
    :rtype: List[str]
    """
    try:
        return smilite.get_zincid_from_smile(smile, backend=backend)
    except URLError:
        ssl._create_default_https_context = ssl._create_unverified_context
        return smilite.get_zincid_from_smile(smile, backend=backend)


def batch_get_smiles_from_zinc(
    zinc_ids: List[str], backend: str = "zinc15"
) -> Dict[str, str]:
    """Gets the corresponding SMILE string(s) for a list of ZINC IDs.

    :param zinc_ids: List of ZINC IDs, e.g., ``['ZINC00029323', 'ZINC00029324']``
    :type zinc_ids: List[str]
    :return: _description_
    :rtype: Dict[str, str]
    """
    return {
        zinc_id: get_smiles_from_zinc(zinc_id, backend=backend)
        for zinc_id in tqdm(zinc_ids)
    }


def batch_get_zinc_id_from_smiles(
    smiles: List[str], backend: str = "zinc15"
) -> Dict[str, List[str]]:
    """
    Gets the corresponding ZINC ID for a list of smile string queries from
    the ZINC online database. Requires an internet connection.

    :param smile_str: A list of valid SMILE string, e.g.,
        ``["C[C@H]1CCCC[NH+]1CC#CC(c2ccccc2)(c3ccccc3)O", "CCC"]``
    :type smile_str: str
    :param backend:  Specifies the database backend, ``"zinc12"`` or ``"zinc15"``.
    :type backend: str
    :returns: the SMILE string for the corresponding ZINC ID(s) in a list.
    :rtype: Dict[str, List[str]]
    """
    return {
        s: get_zinc_id_from_smile(s, backend=backend) for s in tqdm(smiles)
    }



"""Command line interface for graphein."""

import pathlib
import pickle

import rich_click as click

from graphein import __version__
from graphein.protein.graphs import construct_graph
from graphein.utils.config_parser import parse_config


@click.command()
@click.version_option(__version__)
@click.option(
    "-c",
    "--config_path",
    help="The name the of .yml input config",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path
    ),
)
@click.option(
    "-p",
    "--path",
    help="Path to input pdbs",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=True, path_type=pathlib.Path
    ),
)
@click.option(
    "-o",
    "--output_path",
    help="Path to output dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=pathlib.Path
    ),
)
def main(config_path, path, output_path):
    """Build the graphs and save them in output dir."""
    config = parse_config(path=config_path) if config_path else None
    if path.is_file():
        paths = [path]
    elif path.is_dir():
        paths = list(path.glob("*.pdb"))
    else:
        raise NotImplementedError(
            "Given PDB path needs to point to either a pdb file or a directory with pdb files."
        )

    for path in paths:
        g = construct_graph(config=config, path=str(path))

        with open(str(output_path / f"{path.stem}.gpickle"), "wb") as f:
            pickle.dump(g, f)


if __name__ == "__main__":
    main()



import os
import sys
from functools import wraps
from shutil import which
from typing import Optional


class MissingDependencyError(Exception):
    """Raised when a required dependency is missing."""

    def __init__(self, message):
        super().__init__(message)
        self.message = message


def import_message(
    submodule: str,
    package: str,
    conda_channel: Optional[str] = None,
    pip_install: bool = False,
    extras: bool = False,
) -> str:
    """
    Return warning if package is not found.
    Generic message for indicating to the user when a function relies on an
    optional module / package that is not currently installed. Includes
    installation instructions. Typically used in conjunction without optional featurisation libraries

    :param submodule: graphein submodule that needs an external dependency.
    :type submodule: str
    :param package: External package this submodule relies on.
    :type package: str
    :param conda_channel: Conda channel package can be installed from, if at all. Defaults to ``None``.
    :type conda_channel: str, optional
    :param pip_install: Whether package can be installed via pip. Defaults to ``False``.
    :type pip_install: bool
    :param extras: Whether package would be installed with
        ``pip install graphein[extras]``. Defaults to ``False``.
    :type extras: bool
    """
    is_conda = os.path.exists(os.path.join(sys.prefix, "conda-meta"))
    installable = True
    if is_conda:
        if conda_channel is None:
            installable = False
            installation = f"{package} cannot be installed via conda"
        else:
            installation = f"conda install -c {conda_channel} {package}"
    elif pip_install:
        installation = f"pip install {package}"
    else:
        installable = False
        installation = f"{package} cannot be installed via pip"

    message = f"To use the Graphein submodule {submodule}, you need to install: {package} "
    if installable:
        message += f"\nTo do so, use the following command: {installation}"
    else:
        message += f"\n{installation}"

    if extras:
        message += (
            "\nAlternatively, you can install graphein with the extras: "
        )
        message += "\n\npip install graphein[extras]"
    return message


def is_tool(name: str, error: bool = False) -> bool:
    """Checks whether ``name`` is on ``PATH`` and is marked as an executable.

    Source:
    https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    :param name: Name of program to check for execution ability.
    :type name: str
    :param error: Whether to raise an error.
    :type error: bool. Defaults to ``False``.
    :return: Whether ``name`` is on PATH and is marked as an executable.
    :rtype: bool
    :raises MissingDependencyError: If ``error`` is ``True`` and ``name`` is
        not on ``PATH`` or is not marked as an executable.
    """
    found = which(name) is not None
    if not found and error:
        raise MissingDependencyError(name)
    return found


# Decorator for checking if a function has the required dependencies
def requires_external_dependencies(*deps):
    """
    A decorator to check if all required dependencies are installed before
    calling the decorated function. If a dependency is missing, it raises
    a MissingDependencyError.

    :param deps: A list of dependencies (as strings) to check for.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            missing_deps = [dep for dep in deps if not is_tool(dep)]
            if missing_deps:
                missing = ", ".join(missing_deps)
                raise MissingDependencyError(
                    f"Missing dependencies: {missing}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


_lib_check_cache = {}


def requires_python_libs(*libs):
    """
    A decorator to check if all required Python library dependencies are installed
    before calling the decorated function. If a library is missing, it raises
    an ImportError with details about the missing libraries. Caches check results
    to avoid repeated imports.

    :param libs: A list of library names (as strings) to check for.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            missing_libs = []
            for lib in libs:
                # Check if the library check result is cached
                if lib in _lib_check_cache:
                    if not _lib_check_cache[lib]:
                        missing_libs.append(lib)
                else:
                    try:
                        __import__(lib)
                        _lib_check_cache[lib] = True
                    except ImportError:
                        _lib_check_cache[lib] = False
                        missing_libs.append(lib)
            if missing_libs:
                missing = ", ".join(missing_libs)
                raise ImportError(
                    f"Missing Python library dependencies: {missing}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""Functions for featurising Small Molecule Graphs.

Plotting functions for molecules wrap the methods defined on protein graphs and provide sane defaults.
"""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from typing import Optional, Tuple

import matplotlib.pyplot as plt
import networkx as nx

from graphein.protein.visualisation import (
    plot_protein_structure_graph,
    plotly_protein_structure_graph,
)


def plot_molecular_graph(
    G: nx.Graph,
    angle: int = 30,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (10, 7),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 1,
    label_node_ids: bool = True,
    node_colour_map: plt.cm = plt.cm.plasma,
    edge_color_map: plt.cm = plt.cm.plasma,
    colour_nodes_by: str = "element",
    colour_edges_by: str = "kind",
    edge_alpha: float = 0.5,
    plot_style: str = "ggplot",
    out_path: Optional[str] = None,
    out_format: str = ".png",
) -> Axes3D:
    """
    Plots molecular graph in ``Axes3D``.

    :param G:  nx.Graph Protein Structure graph to plot.
    :type G: nx.Graph
    :param angle:  View angle. Defaults to ``30``.
    :type angle: int
    :param plot_title: Title of plot. Defaults to ``None``.
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to ``(10, 7)``.
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to ``0.7``.
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size, defaults to ``20``.
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes reflect degree. Defaults to ``20``.
    :type node_size_multiplier: float
    :param label_node_ids: bool indicating whether or not to plot ``node_id`` labels. Defaults to ``True``.
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes. Defaults to ``plt.cm.plasma``.
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges. Defaults to ``plt.cm.plasma``.
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. ``"degree"``, ``"seq_position"`` or a node feature.
    :type colour_nodes_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only ``"kind"`` is supported.
    :type colour_edges_by: str
    :param edge_alpha: Controls edge transparency. Defaults to ``0.5``.
    :type edge_alpha: float
    :param plot_style: matplotlib style sheet to use. Defaults to ``"ggplot"``.
    :type plot_style: str
    :param out_path: If not none, writes plot to this location. Defaults to ``None`` (does not save).
    :type out_path: str, optional
    :param out_format: Fileformat to use for plot
    :type out_format: str
    :return: matplotlib Axes3D object.
    :rtype: Axes3D
    """

    return plot_protein_structure_graph(
        G,
        angle=angle,
        plot_title=plot_title,
        figsize=figsize,
        node_alpha=node_alpha,
        node_size_min=node_size_min,
        node_size_multiplier=node_size_multiplier,
        label_node_ids=label_node_ids,
        node_colour_map=node_colour_map,
        edge_color_map=edge_color_map,
        colour_nodes_by=colour_nodes_by,
        colour_edges_by=colour_edges_by,
        edge_alpha=edge_alpha,
        plot_style=plot_style,
        out_path=out_path,
        out_format=out_format,
    )


def plotly_molecular_graph(
    g: nx.Graph,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (620, 650),
    node_alpha: float = 0.7,
    node_size_min: float = 20,
    node_size_multiplier: float = 1.0,
    label_node_ids: bool = True,
    node_color_map: plt.cm = plt.cm.plasma,
    edge_color_map: plt.cm = plt.cm.plasma,
    colour_nodes_by: str = "element",
    colour_edges_by: str = "kind",
) -> go.Figure:
    """
    Plots molecular graph using plotly.

    :param G:  nx.Graph Molecular graph to plot
    :type G: nx.Graph
    :param plot_title: Title of plot, defaults to ``None``.
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to ``(620, 650)``.
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to ``0.7``.
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size. Defaults to ``20.0``.
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes reflect degree. Defaults to ``1.0``.
    :type node_size_multiplier: float
    :param label_node_ids: bool indicating whether or not to plot ``node_id`` labels. Defaults to ``True``.
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes. Defaults to ``plt.cm.plasma``.
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges. Defaults to ``plt.cm.plasma``.
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. ``"degree"``, or a node feature. Defaults to ``"element"``.
    :type colour_edges_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only ``"kind"`` is supported.
    :type colour_nodes_by: str
    :returns: Plotly Graph Objects plot
    :rtype: go.Figure
    """
    return plotly_protein_structure_graph(
        g,
        plot_title=plot_title,
        figsize=figsize,
        node_alpha=node_alpha,
        node_size_min=node_size_min,
        node_size_multiplier=node_size_multiplier,
        label_node_ids=label_node_ids,
        node_colour_map=node_color_map,
        edge_color_map=edge_color_map,
        colour_nodes_by=colour_nodes_by,
        colour_edges_by=colour_edges_by,
    )
